# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Abstraction for quantizing neural networks implemented in jax."""

import contextlib
import dataclasses
import enum
import functools
import logging
import typing
from typing import Iterable, Optional, Tuple, Union

from absl import flags
from aqt.jax_legacy.jax import compute_cost_utils
from aqt.jax_legacy.jax import fp_cast
from aqt.jax_legacy.jax import get_bounds
from aqt.jax_legacy.jax import primitives
from aqt.jax_legacy.jax import shape_utils
from aqt.jax_legacy.jax import utils
from aqt.jax_legacy.jax.flax import struct as flax_struct
from flax import linen as nn
import jax
from jax import lax
import jax.numpy as jnp

# Global bool to control the use of epsilon in the denominator of the scaling
# methods signed_int_scale and unsigned_int_scale. Epsilon is added to avoid
# division by 0. For testing, one may choose to disable the epsilon by setting
# this global to True.
# As this is a global variable, please modify it only before calling any
# functions that use it.
DISABLE_EPSILON_IN_SCALE_FUN_FOR_TESTING = False

# Dtype for quantization computations: scaling; floor and clip; rescaling. this
# is chosen to optimize performance for a given hardware i.e. for TPU we set it
# to float32. It should be matching native dtype of the hardware's
# 'vector unit'.
SCALE_DTYPE = jnp.float32

dataclass = (
    flax_struct.dataclass if not typing.TYPE_CHECKING else dataclasses.dataclass
)

# ActBounds can be an Jax array of floats with a shape that is broadcastable to
# the shape of activation tensors.
ActsBoundT = Union[
    float,
    jnp.ndarray,
    get_bounds.GetBounds.Hyper,
    get_bounds.DynamicBounds.Hyper,
    None,
]


@dataclass
class _FloatQuant:
  """Parameters for floating-point quantization.

  Floating-point quantization refers to degraded floating-point precision
  below those natively supported, e.g., bfloat16. This quantization scheme
  can either work with, or without scaling (controlled by `is_scaled`).

  With scaling, these quantization steps follow,

  1. Use the maximum representable floating-point value to determine a scale.
  2. This scale is used to "upscale" the argument to the range of the target
     floating-point format.
  3. The precision of the argument is then degraded through a downcast
     operation.
  4. Finally the degraded-precision result is "downscaled" by the inverse
     scale.

  Without scaling, these quantization steps follow,

  1. The argument is downcast to the target fp-format with degraded precision.
     Of importance in this downcast is the saturating behavior, which is
     logically equivalent to clipping by the maximum representable target
     value.
  """

  @dataclass
  class FloatPrec:
    """Parameters for specifying a custom precision floating-point type."""

    # The minimum exponent value of the quantized floating-point format.
    exp_min: int
    # The maximum exponent value of the quantized floating-point format.
    exp_max: int
    # The number of significand bits (excluding hidden bit) of the quantized
    # floating-point format.
    sig_bits: int

  # Whether or not floating-point fake-quant makes use of scaling.
  is_scaled: bool
  # Precision specification for floating-point quantization.
  fp_spec: FloatPrec


_PrecT = Union[None, int, _FloatQuant]  # pylint: disable=invalid-name


class QuantType(str, enum.Enum):
  """Quantization strategy dataclass."""

  # fake_quant strategy ensures that quantized values form an arithmetic
  # sequence e.g.  0*s ... 255*s for 8-bit positive quantization, for some s.
  # it can be implemented as a local op: upscale, floor, clip, downscale.
  FAKE_QUANT = 'fake_quant'
  # fake_quant strategy with quantized inputs/weights type-casted to int.
  FAKE_QUANT_WITH_INT = 'fake_quant_with_int'
  # aqt ensures that MatMul/Conv are in actual integer domain.
  # It can't be implemented as a single op.
  # Before matmul we have upscale, floor and clip, and after matmul we have
  # downscale.
  AQT = 'aqt'

  def to_jax_type(self) -> primitives.jnp_dtype:
    """Returns quantized dtype for the corresponding quantization strategy."""
    # Currently, this function is used to decide the return type for
    # 'QuantOps.to_quantized.' The AQT implementation works by having a
    # conversion to an int dtype and then  back to a fp dtype happen *within*
    # to_quantized, so that Jax backprop works correctly. Thus
    # counter-intuitively, we need this to return a fp dtype for 'aqt' since the
    # return type for 'to_quantized' overall is fp. TODO(malmaud): As part of
    # the refactor of this module, clean this up to eliminate the
    # counter-intuitive behavior.
    if self.value in ['aqt', 'fake_quant']:  # pylint: disable=comparison-with-callable
      return SCALE_DTYPE
    elif self.value == 'fake_quant_with_int':  # pylint: disable=comparison-with-callable
      return jnp.int8
    else:
      raise RuntimeError(f'QuantType {self.value} is unknown.')


class QuantOps:
  """Class for quantizing and dequantizing weights and activations."""

  # Redefined here as nested class attributes to avoid forward-references.
  FloatQuant = _FloatQuant  # pylint: disable=invalid-name
  PrecT = _PrecT  # pylint: disable=invalid-name

  @dataclass
  class WeightParams:
    """Parameters for weight quantization."""

    prec: _PrecT  # expected precision for weight quantization.
    # enable all available values during quantization
    half_shift: bool
    # Axis along which to quantize weights (the non-feature axis).
    axis: Optional[Iterable[int]]


    # expected scale shape for weights quantization. Defaults to None.
    expected_scale_shape: Union[None, int, Tuple[int, ...]] = None

  @dataclass
  class ActHParams:
    """Parameters for activation quantization."""

    # Inheriting from 'str' and making the enums have string values lets us
    # conveniently serialize this class to JSON without a custom JSON encoder.
    class InputDistribution(str, enum.Enum):
      SYMMETRIC = 'symmetric'
      POSITIVE = 'positive'

    input_distribution: InputDistribution
    # float means fixed bound. '-1' means no quantization.
    bounds: ActsBoundT
    prec: _PrecT
    half_shift: bool


  def __init__(
      self,  #
      *,
      prec: Union[int, _FloatQuant],
      scale: Optional[jnp.ndarray],
      symmetric: bool,
      bounds: Optional[jnp.ndarray],
      half_shift: bool,
  ):
    """Default constructor, use of named constructors is strongly encoraged.

    Args:
      prec: precision for the QuantOps
      scale: scaling factor to scale the input to quantized precision range
      symmetric: whether the input to quantize is symmetric
      bounds: Optional. The clipping bounds used for calculating scale factors.
      half_shift: Symmetric quantization with all available values enabled
    """
    self._prec = prec
    self._half_shift = half_shift
    if scale is None:
      self._scale = None
    else:
      self._scale = scale.astype(SCALE_DTYPE)

    self._symmetric = symmetric
    # Storing bounds are useful for two reasons: one is debugging, since it
    # makes easy to see how a QuantOps instance came up with its scale factor.
    # Two is that right now, we use a bounds of '-1' as a special value meaning
    # to 'not quantize'. See comment on the 'should_quantize' method for more
    # details.
    self._bounds = bounds

  @classmethod
  def create_symmetric_fp(
      cls,
      *,
      bounds: Optional[primitives.BoundsT],
      fp_quant: _FloatQuant,
  ) -> 'QuantOps':
    """Create QuantOps for symmetric clipping to floating-point bounds.

    Args:
      bounds: The upper (and absolute lower) bound to clip the inputs.
      fp_quant: quantization floating-point specification of the target format.

    Returns:
      QuantOps for quantizing/dequantizing signed activations.
    """
    if bounds is None:
      if fp_quant.is_scaled:
        raise ValueError(
            'bounds can only be None if fp_quant.is_scaled is False.'
        )
      return cls(
          prec=fp_quant,
          scale=None,
          symmetric=True,
          bounds=None,
          half_shift=False,
      )  # disable half_shift for fp quantization
    else:
      initial_bounds = bounds
      # We set bounds = -1 to indicate no quantization.
      # TODO(shivaniagrawal): Move away from the hack of setting bound as -1.
      bounds = jnp.asarray(bounds, SCALE_DTYPE)
      if not DISABLE_EPSILON_IN_SCALE_FUN_FOR_TESTING:
        # to avoid log2(0)
        bounds = jnp.abs(bounds) + jnp.finfo(SCALE_DTYPE).eps
      scale = jnp.exp2(-jnp.floor(jnp.log2(bounds)))  # Scale to unit binade.
      # NOTE: stop_gradient is needed here to prevent gradient flow through
      # scale when scale is not a constant, but computed as a function of
      # activations or weights.
      scale = lax.stop_gradient(scale)

      return cls(
          prec=fp_quant,
          scale=scale,
          symmetric=True,
          bounds=initial_bounds,
          half_shift=False,
      )  # disable half_shift for fp quantization

  @classmethod
  def create_symmetric(
      cls, *, bounds: primitives.BoundsT, prec: int, half_shift: bool
  ) -> 'QuantOps':
    """Create QuantOps for symmetric activations clipped to [-bounds, bounds].

    Args:
      bounds: The upper (and absolute lower) bound to clip the inputs.
      prec: Signed int precision for the QuantOps.
      half_shift: Symmetric quantization with all available values enabled

    Returns:
      QuantOps for quantizing/dequantizing signed activations.
    """
    initial_bounds = bounds
    bounds = jnp.asarray(bounds, SCALE_DTYPE)
    if not DISABLE_EPSILON_IN_SCALE_FUN_FOR_TESTING:
      bounds += jnp.finfo(SCALE_DTYPE).eps  # to avoid div by 0
    scale = (
        primitives.signed_int_bound(prec=prec, half_shift=half_shift) / bounds
    )
    # NOTE: stop_gradient is needed here to prevent gradient flow through scale
    # when scale is not a constant, but computed as a function of activations or
    # weights.
    scale = lax.stop_gradient(scale)
    return cls(  # pytype: disable=wrong-arg-types  # jax-ndarray
        prec=prec,
        scale=scale,
        symmetric=True,
        bounds=initial_bounds,
        half_shift=half_shift,
    )

  @classmethod
  def create_positive(
      cls, *, bounds: primitives.BoundsT, prec: int
  ) -> 'QuantOps':
    """Create QuantOps for positive activations clipped to [0, bounds].

    Args:
      bounds: The upper bound to clip the activations.
      prec: Unsigned int precision for the QuantOps.

    Returns:
      QuantOps for quantizing/dequantizing unsigned activations.
    """
    initial_bounds = bounds
    bounds = jnp.asarray(bounds, SCALE_DTYPE)
    if not DISABLE_EPSILON_IN_SCALE_FUN_FOR_TESTING:
      bounds += jnp.finfo(SCALE_DTYPE).eps  # to avoid div by 0
    scale = primitives.unsigned_int_bound(prec=prec) / bounds
    # NOTE: stop_gradient is needed here to prevent gradient flow through scale
    # when scale is not a constant, but computed as a function of activations.
    scale = lax.stop_gradient(scale)
    return cls(  # pytype: disable=wrong-arg-types  # jax-ndarray
        prec=prec,
        scale=scale,
        symmetric=False,
        bounds=initial_bounds,
        half_shift=False,
    )  # disable half_shift for positive distribution

  def assert_scale_shape_is(self, *, shape: Union[int, Tuple[int, ...]]):
    if self._scale is None:
      raise ValueError('scale is None')
    # TODO(shivaniagrawal): add option for float scale for fixed bound acts
    # quantization.
    assert (
        self._scale.shape == shape
    ), 'scale shape is unexpected, should be %s but got %s' % (
        shape,
        self._scale.shape,
    )

  def to_quantized(
      self, x: jnp.ndarray, *, dtype: primitives.jnp_dtype
  ) -> jnp.ndarray:
    """Quantizes the argument to the target format.

    integer: "upscales", rounds or floors and clips.
    floating-point: optionally upscales, then downcasts to target precision.

    Args:
      x: Argument to be quantized.
      dtype: Type of returned quantized value of x. If quantized x is an input
        to a matmul, we might be want to set it to jnp.int8. If quantized x is
        weights stored in memory, same applies. In fake_quant style we might
        prefer to set dtype=SCALE_DTYPE, since quantized x might get constant
        folded with rescale op (`from_quantized`). Please take a look at the
        comment on SCALE_DTYPE.

    Returns:
      Quantized value of x.
    """
    if isinstance(self._prec, _FloatQuant):
      if self._prec.is_scaled:
        x = jnp.multiply(x, self._scale).astype(x.dtype)
      fp_spec = self._prec.fp_spec
      return fp_cast.downcast_sat_ftz(
          x,
          fp_spec.exp_min,
          fp_spec.exp_max,
          fp_spec.sig_bits,
      )
    else:
      if self._symmetric:
        quantize = primitives.round_and_clip_to_signed_int
      else:
        quantize = primitives.floor_and_clip_to_unsigned_int
      scaled_x = jnp.multiply(x, self._scale)
      return quantize(
          scaled_x, prec=self._prec, dtype=dtype, half_shift=self._half_shift
      )

  # Same as to_quantized but it just "downscales" using the same scale.
  def from_quantized(
      self, x: jnp.ndarray, *, dtype: primitives.jnp_dtype
  ) -> jnp.ndarray:
    """'Rescales' the quantized value.

    Args:
      x: quantized.
      dtype: return type for rescaled x

    Returns:
      Rescaled x cast to type dtype
    """

    if isinstance(self._prec, _FloatQuant) and not self._prec.is_scaled:
      return x
    rescaled_x = jnp.divide(x, self._scale)
    return rescaled_x.astype(dtype)

  # Helper fake quantization
  def fake_quant(
      self,
      x: jnp.ndarray,
      *,
      quantized_type: primitives.jnp_dtype,
      fake_dependency: Optional[jnp.ndarray] = None,
  ) -> jnp.ndarray:
    del fake_dependency  # unused; this was a remnant of pre-omnistaging JAX.
    x_dtype = x.dtype
    quantized_x = self.to_quantized(x, dtype=quantized_type)
    return self.from_quantized(quantized_x, dtype=x_dtype)

  # Assumes weights are unsigned int of precision prec.
  @classmethod
  def create_weights_ops(
      cls,
      w: jnp.ndarray,
      *,
      weight_params: WeightParams,
  ) -> 'QuantOps':
    """Create a QuantOps that can quantize and dequantize a weight tensor.

    Args:
      w: The weights to quantize.
      weight_params: WeightParams Parameters required for weight quantization.

    Returns:
      Quantized and rescaled inputs using fake quant approach.
    """
    weight_bounds = primitives.max_abs_weights(w, axis=weight_params.axis)
    prec = weight_params.prec
    half_shift = weight_params.half_shift
    if isinstance(prec, _FloatQuant):
      ops = cls.create_symmetric_fp(bounds=weight_bounds, fp_quant=prec)
    else:
      ops = cls.create_symmetric(
          bounds=weight_bounds, prec=prec, half_shift=half_shift
      )


    if weight_params.expected_scale_shape is not None:
      # NOTE: We set keepdim to True when computing weights scale, as a result
      # the axes which are reduced are left in the result as dimensions with
      # size one. User should correctly pass the shape with reduced dimensions
      # set to 1.
      ops.assert_scale_shape_is(shape=weight_params.expected_scale_shape)
    return ops

  # Assumes weights are unsigned int of precision prec.
  @classmethod
  def create_weights_fake_quant(
      cls,
      w: jnp.ndarray,
      *,
      weight_params: WeightParams,
      quantized_type: primitives.jnp_dtype = SCALE_DTYPE,
      fake_dependency: Optional[jnp.ndarray] = None,
      quantize_weights: bool = True,
  ) -> jnp.ndarray:
    """Quantize weights with fake quant approach.

    Args:
      w: The weights to quantize.
      weight_params: WeightParams Parameters required for weight quantization.
      quantized_type: type of intermediate quantized value of weights. Defaults
        to SCALE_DTYPE.
      fake_dependency: dynamic array, quantized weights will have fake
        dependency on. lax.tie_in for more details. This is used in order to
        prevent constant folding of rescale op with quantized weights. Defaults
        to None, in this case  quantized weights would not have a fake
        dependency.
      quantize_weights: whether weights should be quantized or not

    Returns:
      Quantized and rescaled inputs using fake quant approach.
    """
    # TODO(yichi): if weight_params.prec is None or weight_binarize flag True:
    if weight_params.prec is None or not quantize_weights:
      return w

    weight_shape = w.shape


    ops = cls.create_weights_ops(w, weight_params=weight_params)
    weight_rescaled = ops.fake_quant(
        w, quantized_type=quantized_type, fake_dependency=fake_dependency
    )


    return weight_rescaled

  # TODO(malmaud): rename 'input' to activation here and elsewhere in this file.
  @classmethod
  def create_input_ops(
      cls,
      inputs: jnp.ndarray,
      *,
      hparams: ActHParams,
      bounds_params: Union[
          get_bounds.GetBounds.Params, get_bounds.DynamicBounds.Params, None
      ],
  ) -> 'QuantOps':
    """Create a QuantOps that can quantize and dequantize an activation tensor.

    Args:
      inputs: The inputs to quantize.
      hparams: Input hyperparameter (ActHParams).
      bounds_params: BoundsParams. Parameters for GetBounds or DynamicBounds.

    Returns:
      Quantized and rescaled inputs using fake quant approach.
    """

    # TODO(shivaniagrawal): investigate why pytype allows types other than
    # ActsBoundT.
    if isinstance(hparams.bounds, int):
      hparams.bounds = float(hparams.bounds)

    # NOTE: if flax module name is None, default name is used.

    # If we want to train with no quantization at first and then turn on
    # GetBounds quantization, we still have to call GetBounds even before
    # quantization is enabled since GetBounds calculates and stores the running
    # statistics that we will use once quantization is enabled. But before
    # quantization is enabled, we want to ignore the returned bounds and just
    # return the original unquantized input. To do so, we take advantage of the
    # fact that GetBounds returns a constant fixed bound for an initial time
    # period and set that initial bound to a special value (-1) to indicate we
    # want to store activation statistics without applying quantization. That
    # will cause clip_bounds will be a tensor of all '-1', which we will check
    # for in a lax.cond call below.

    # TODO(malmaud): Refactor code to separate bounds calculation from tracking
    # activation statistics to avoid the need to rely on special bounds values
    # when disabling quantization.
    if isinstance(hparams.bounds, get_bounds.GetBounds.Hyper):
      if not bounds_params and not isinstance(
          bounds_params, get_bounds.GetBounds.Params
      ):
        raise ValueError(
            'act_hparams.bounds is of type GetBounds.Hyper, user must '
            'provide bounds_params, parameters for GetBounds.'
        )
      assert bounds_params is not None
      clip_bounds = get_bounds.GetBounds(
          hyper=hparams.bounds, name=bounds_params.module_name
      )(
          inputs,
          bounds_params=bounds_params,
      )
    elif isinstance(hparams.bounds, get_bounds.DynamicBounds.Hyper):
      if not bounds_params and not isinstance(
          bounds_params, get_bounds.DynamicBounds.Params
      ):
        raise ValueError(
            'act_hparams.bounds is of type DynamicBounds.Hyper, user must '
            'provide bounds_params, parameters for DynamicBounds.'
        )
      assert bounds_params is not None
      clip_bounds = get_bounds.DynamicBounds(
          hyper=hparams.bounds, name=bounds_params.module_name
      )(
          inputs,
          bounds_params=bounds_params,
      )

    elif isinstance(hparams.bounds, (float, jnp.ndarray)):
      clip_bounds = hparams.bounds
    else:
      assert False, (
          '%s is not a valid type for hparams.bounds, should be float, a list '
          'of floats, DynamicBounds.Hyper or GetBounds.Hyper.'
          % (type(hparams.bounds))
      )

    if isinstance(hparams.prec, _FloatQuant):
      ops = cls.create_symmetric_fp(bounds=clip_bounds, fp_quant=hparams.prec)
    elif (
        hparams.input_distribution == cls.ActHParams.InputDistribution.SYMMETRIC
    ):
      ops = cls.create_symmetric(
          bounds=clip_bounds, prec=hparams.prec, half_shift=hparams.half_shift
      )
    elif (
        hparams.input_distribution == cls.ActHParams.InputDistribution.POSITIVE
    ):
      ops = cls.create_positive(bounds=clip_bounds, prec=hparams.prec)
    else:
      assert False, "can't happen."

    if bounds_params and bounds_params.expected_bounds_shape is not None:
      if isinstance(hparams.bounds, get_bounds.GetBounds.Hyper):
        ops.assert_scale_shape_is(shape=bounds_params.expected_bounds_shape)
      else:
        logging.info(
            'Ignoring value of argument expected_scale_shape. Scale for fixed '
            'bounds would be scalar.'
        )
    return ops

  @classmethod
  def create_inputs_fake_quant(
      cls,
      inputs: jnp.ndarray,
      *,
      hparams: ActHParams,
      bounds_params: Union[
          None, get_bounds.GetBounds.Params, get_bounds.DynamicBounds.Params
      ],
  ) -> jnp.ndarray:
    """Quantize input with fake quant approach.

    Args:
      inputs: The inputs to quantize.
      hparams: Input hyperparameter (ActHParams).
      bounds_params: GetBounds or DynamicBounds Params.

    Returns:
      Quantized and rescaled inputs using fake quant approach.
    """

    if hparams.bounds is None or hparams.prec is None:
      # TODO(lew): support bound-clipping without quantization
      return inputs

    ops = cls.create_input_ops(
        inputs, hparams=hparams, bounds_params=bounds_params
    )

    quantized_inputs = ops.fake_quant(inputs, quantized_type=SCALE_DTYPE)
    return lax.cond(
        ops.should_quantize(),
        lambda _: quantized_inputs,
        lambda _: inputs,
        None,
    )

  # When using GetBounds quantization (if hparams.bounds is an instance of
  # GetBounds.Hyper), if we want to disable quantization but continue to
  # collect activation statistics, we have GetBounds return a clip_bounds
  # tensor to all '-1' values as a signal that quantization shoulnd't be
  # applied. See comment on the call to 'GetBounds' above.
  # TODO(malmaud): Find a less hacky way to do this.
  def should_quantize(self) -> jnp.ndarray:
    """Return whether QuantOps should quantize."""
    # We return a scalar jnp.ndarray of dtype bool instead of a Python bool
    # because during the Jax JIT compilation, self._bounds will be a tracer
    # instead of a concrete tensor, which can't be coerced to a Python bool.
    # Since the type of jnp.all is an ndarray, we invert it with '~' instead of
    # 'not'
    return ~jnp.all(self._bounds == -1)

  def get_scale_for_aqt(self, *, allow_per_channel_scales: bool) -> jnp.ndarray:
    """Returns the scale in a shape appropriate for AQT.

    An error is raised if the granularity of the scale factors are incompatible
    with the current AQT implementation and the setting of
    'allow_per_channel_scales'.

    Args:
      allow_per_channel_scales: A boolean indicating whether a separate scale
        factor is allowed for each output channel (True) or if only a scalar
        (ie, per-layer) scale factor is allowed (False).

    Returns:
      Either a scalar array that correspond to a per-layer scale factor, or an
      array of shape (1, num_channels) that correspond to per-channel scale
      factors.
    """

    scale = self._scale
    if scale is None:
      raise ValueError('scale is None.')

    # If 'scale' is a 1x1x...x1 matrix (ie, only has one element), we
    # canonicalize it to a scalar to simplify the shape-handling code in the AQT
    # implementation.
    if scale.size == 1:
      return scale.reshape(())
    # If the caller requested a a single per-layer scaling factor but the scale
    # factor is non-scalar, raise an error.
    del allow_per_channel_scales
    return scale


PrecisionType = typing.Any


def quantized_dot(
    *,
    w: jnp.ndarray,
    act: jnp.ndarray,
    quant_type: QuantType,
    weight_params: QuantOps.WeightParams,
    act_hparams: Optional[QuantOps.ActHParams],
    bounds_params: Union[
        get_bounds.GetBounds.Params, get_bounds.DynamicBounds.Params, None
    ],
    prefer_int8_to_int32_dot: bool,
    dot_precision: Optional[PrecisionType] = None,
) -> jnp.ndarray:
  """LAX dot with optionally quantized weights and activations.

  Wraps LAX's `Dot
  <https://github.com/google/jax/blob/f65a327c764406db45e95048dfe09209d8ef6d37/jax/_src/lax/lax.py#L632`_
  operator.

  Args:
    w: an array representing weights
    act: an array representing activations
    quant_type: quantization strategy
    weight_params: QuantOps.WeighstParams instance for describing weights
      quantization.
    act_hparams: Optional activation quantization hyperparamers; instance of
      QuantOps.ActHParams. None would mean no activation quantization.
    bounds_params: Optional bounds params for auto activation quantization;
      instance of GetBounds.Params or DynamicBounds.Params.
    prefer_int8_to_int32_dot:  Whether to feed lax.dot inputs with an int8 dtype
      and accumulate to int32 dtype if quantizing to 8bits or 4bits. If False,
      inputs are always foating-point.
    dot_precision: Optional. Either ``None``, which means the default precision
      for the backend, or a ``lax.Precision`` enum value (``Precision.DEFAULT``,
      ``Precision.HIGH`` or ``Precision.HIGHEST``).

  Returns:
    An array containing the result with the same dtype as 'w' and 'act'.

  Raises:
    RuntimeError: 'quant_type' had an unrecognized value.
    TypeError: 'act' and 'w' has different input types.
    ValueError: Shapes of 'act' and 'w' not compatible with quant_type.
  """
  # This code was initially expanded from
  # https://github.com/google/jax/blob/f65a327c764406db45e95048dfe09209d8ef6d37/jax/_src/lax/lax.py#L632
  # We keep the original return-value semantics of lax.dot, which this wraps. In
  # particular, the type of the return value of quantized_dot is the same as the
  # type of the inputs. That means that if the inputs are bfloat16, then the
  # return type of this function will also be bfloat16 even though on current
  # TPUs the underlying bf16*bf16 matrix-multiplication accumulates results to
  # float32. This is potentially undesirable since the user might want the raw
  # float32 result, but it ultimately stems from a limitation of the HLO 'dot'
  # instruction. If that instruction updates to support user-specified output
  # types, we could update quantized_dot accordingly to take a dtype argument to
  # control the return value type. This applies equally to
  # quantized_dynamic_dot_general.
  if not (
      1 <= act.ndim <= 2 and 1 <= w.ndim <= 2 and act.shape[-1] == w.shape[0]
  ):
    raise ValueError(
        'Incompatible shapes for dot: got {} and {}.'.format(act.shape, w.shape)
    )
  dot_dimension_numbers = (((act.ndim - 1,), (0,)), ((), ()))
  return quantized_dot_general(
      w=w,
      act=act,
      quant_type=quant_type,
      weight_params=weight_params,
      act_hparams=act_hparams,
      bounds_params=bounds_params,
      prefer_int8_to_int32_dot=prefer_int8_to_int32_dot,
      dimension_numbers=dot_dimension_numbers,
      dot_precision=dot_precision,
  )


def _canonicalize_feature_axes(
    axis: Union[int, Tuple[int, ...]], ndims: int
) -> Tuple[int, ...]:
  """Returns feature axes from reduction axes."""
  if not isinstance(axis, tuple):
    axis = (axis,)
  if not all([dim < ndims and dim >= -ndims for dim in axis]):
    raise ValueError(
        f'Axis {axis} should correspond to dimensions of the weight.'
    )
  axis = tuple(sorted([(ndims + dim) % ndims for dim in axis]))
  # Feature axes must be either first len(Feature_axis) consecutive
  # dimensions of weight or last len(axis) consecutive axes.
  is_first_consecutive_dims = axis == tuple(range(len(axis)))
  if not is_first_consecutive_dims:
    raise ValueError(
        f'Reduction axis {axis} should be consecutive first dimensions '
        'of the weight.'
    )
  return axis


@dataclass
class QuantW:
  quantized_w: jnp.ndarray
  scale: jnp.ndarray


def flaxformer_dot_general(
    *,
    act: jnp.ndarray,
    w: Optional[jnp.ndarray],
    dimension_numbers: lax.DotDimensionNumbers,
    weight_params: QuantOps.WeightParams,
    act_hparams: Optional[QuantOps.ActHParams],
    bounds_params: Union[
        None, get_bounds.DynamicBounds.Params, get_bounds.GetBounds.Params
    ],
    dot_precision: Optional[PrecisionType] = None,
    prefer_int8_to_int32_dot: bool = True,
    quant_w: Optional[QuantW] = None,
) -> jnp.ndarray:
  """Flaxformer dot general with optionally quantized weights and activations."""
  input_dtype = act.dtype
  is_weight_quantized = False
  is_act_quantized = False
  (axis, _), (_, _) = dimension_numbers
  act_scale_shape = tuple(
      [act.shape[dim] if dim not in axis else 1 for dim in range(0, act.ndim)]
  )
  if w is not None:
    weight_scale_shape = (1,) * len(axis) + tuple(
        [w.shape[dim] for dim in range(len(axis), w.ndim)]
    )
  elif quant_w is not None:
    weight_scale_shape = (1,) * len(axis) + tuple([
        quant_w.quantized_w.shape[dim]
        for dim in range(len(axis), quant_w.quantized_w.ndim)
    ])
  else:
    raise ValueError('Either of w or quant_w must be not None')

  if act_hparams is not None and act_hparams.prec is not None:
    is_act_quantized = True
    # Calculate 's', the per-column scale factor on activations.
    act_op = QuantOps.create_input_ops(
        act, hparams=act_hparams, bounds_params=bounds_params
    )
    assert act_op._scale is not None  # pylint: disable=protected-access
    act_quantized = act_op.to_quantized(act, dtype=input_dtype)

    # Now calculate s^-1.  First we extract s, the activation scale factor,
    # into a  variable called 'act_scale'. We extract it from 'act_op', the
    # QuantOps instance that calculated the scale factors for the activation
    # matrix.
    act_scale = act_op._scale.astype(input_dtype)  # pylint: disable=protected-access
    if act_scale.ndim == 0:
      act_scale = act_scale * jnp.ones(act_scale_shape, act_scale.dtype)
  else:
    # In this case, activations are not being quantized; only weights. There
    # is no need to absorb activation scales into the rows of the weight
    # matrix so 'w_scaled_rows' can just be set to the original weight matrix.
    act_quantized = act
    act_scale = jnp.ones(act_scale_shape, act.dtype)

  if weight_params is not None and weight_params.prec is not None:
    is_weight_quantized = True
    if quant_w is not None:
      weight_quantized = quant_w.quantized_w.astype(input_dtype)
      weight_scale = quant_w.scale
    else:
      # Calculate 'r' from (s^-1) * w
      weight_op = QuantOps.create_weights_ops(w, weight_params=weight_params)
      weight_scale = weight_op._scale.astype(input_dtype)  # pylint: disable=protected-access

      if weight_params.expected_scale_shape:
        shape_utils.assert_shapes_equal(
            weight_scale.shape, weight_params.expected_scale_shape
        )

      # Quantize weight matrix by calculating RoundAndClip(s^-1 * w * t)
      # TODO(malmaud): See comment on 'act_op.to_quantized' above, which
      # applies here as well.
      weight_quantized = weight_op.to_quantized(w, dtype=input_dtype)
  else:
    assert (
        w is not None
    ), 'w can not be None if weight quantization is not specified.'
    weight_quantized = w
    weight_scale = jnp.ones(weight_scale_shape, w.dtype)

  weight_prec = None if weight_params is None else weight_params.prec
  metadata_context = contextlib.suppress()
  if flags.FLAGS.metadata_enabled:
    metadata_context = compute_cost_utils.DotMetadataMonkeyPatch(
        lhs_prec=None, rhs_prec=weight_prec, rhs_is_weight=True
    )
  # Use metadata context to annotate op metadata with quantization info

  # TODO(shivaniagrawal): this is duplicated code with quantized_dot that can
  # be de-duplicated.
  # To decide whether to use an integer-domain dot operation, we first check
  # if the static quantization parameters are compatible with it by seeing if
  # they request that both inputs be quantized 8bits or less. Then check if
  # the dynamic parameters are compatible with it. ie, in a training run with
  # quantization enabled, are we past the activation start step yet.

  # We also do not use int8_to_int32_dot if activation has positive
  # distribution and prec=8, since we would not be able to fit uint8 range in
  # int8.
  # TODO(shivaniagrawal): A proper solution for this would be to have mixed
  # dot(uint8, int8) -> int32 in XLA.

  act_prec = None if act_hparams is None else act_hparams.prec
  act_has_symm_distribution = act_hparams is not None and (
      act_hparams.input_distribution
      == QuantOps.ActHParams.InputDistribution.SYMMETRIC
  )
  weight_prec = None if weight_params is None else weight_params.prec

  # To decide whether to use an integer-domain dot operation, we first check
  # if the static quantization parameters are compatible with it by seeing if
  # they request that both inputs be quantized 8bits or less. Then check if
  # the dynamic parameters are compatible with it. ie, in a training run with
  # quantization enabled, are we past the activation start step yet.

  # We also do not use int8_to_int32_dot if activation has positive
  # distribution and prec=8, since we would not be able to fit uint8 range in
  # int8.
  # TODO(shivaniagrawal): A proper solution for this would be to have mixed
  # dot(uint8, int8) -> int32 in XLA.
  weight_fits_in_int8 = is_weight_quantized and (
      weight_prec is not None and weight_prec <= 8
  )
  # is_act_quantized might be an instance of a Jax tracer instead of a
  # Python boolean since it is generally computed from a dynamic input to a
  # JITted Jax function. Thus we use '&' instead of 'and'.
  act_prec_fits_int8 = act_prec is not None and (
      (act_prec == 8 and act_has_symm_distribution) or (act_prec < 8)
  )
  act_fits_in_int8 = is_act_quantized & act_prec_fits_int8
  use_int8_to_int32_dot = (
      prefer_int8_to_int32_dot & weight_fits_in_int8 & act_fits_in_int8
  )

  with metadata_context:
    # Calculate matmul(...)
    # we are not quantizing activations yet, hence not using int8 matmul
    out_quantized = dot_general_aqt(
        act_quantized,
        weight_quantized,
        dimension_numbers=dimension_numbers,
        dot_precision=dot_precision,
        use_int8_to_int32_dot=use_int8_to_int32_dot,
    )

  if is_weight_quantized or is_act_quantized:

    inv_scale = lax.dot_general(
        (1 / act_scale),
        (1 / weight_scale),
        dimension_numbers=dimension_numbers,
    )
    return out_quantized * inv_scale
  else:
    return out_quantized


# TODO(shivaniagrawal): extend it for generic dot_dimenson_numbers
def quantized_dot_general(
    *,
    w: Optional[jnp.ndarray],
    act: jnp.ndarray,
    quant_type: QuantType,
    weight_params: QuantOps.WeightParams,
    act_hparams: Optional[QuantOps.ActHParams],
    bounds_params: Union[
        None, get_bounds.GetBounds.Params, get_bounds.DynamicBounds.Params
    ],
    prefer_int8_to_int32_dot: bool,
    dimension_numbers: lax.DotDimensionNumbers,
    dot_precision: Optional[PrecisionType] = None,
    quant_w: Optional[QuantW] = None,
) -> jnp.ndarray:
  """LAX dot general with optionally quantized weights and activations.

  Wraps LAX's `Dot General
  <https://jax.readthedocs.io/en/latest/_autosummary/jax.lax.dot_general.html.

  Args:
    w: an array representing weights
    act: an array representing activations
    quant_type: quantization strategy
    weight_params: QuantOps.WeighstParams instance for describing weights
      quantization.
    act_hparams: Optional activation quantization hyperparamers; instance of
      QuantOps.ActHParams. None would mean no activation quantization.
    bounds_params: Optional bounds params for auto activation quantization;
      instance of GetBounds.Params or DynamicBounds.Params.
    prefer_int8_to_int32_dot:  Whether to feed lax.dot inputs with an int8 dtype
      and accumulate to int32 dtype if quantizing to 8bits or 4bits. If False,
      inputs are always foating-point.
    dimension_numbers: a tuple of tuples of the form `((lhs_contracting_dims,
      rhs_contracting_dims), (lhs_batch_dims, rhs_batch_dims))`.
    dot_precision: Optional. Either ``None``, which means the default precision
      for the backend, or a ``lax.Precision`` enum value (``Precision.DEFAULT``,
      ``Precision.HIGH`` or ``Precision.HIGHEST``).
    quant_w: Quantization weights and scales, provided as class QuantW. If
      provided, quantization variables are used directly. Defaults to None.

  Returns:
    An array containing the result with the same dtype as 'w' and 'act'.

  Raises:
    RuntimeError: 'quant_type' had an unrecognized value.
    TypeError: 'act' and 'w' has different input types.
    ValueError: Shapes of 'act' and 'w' not compatible with quant_type.
  """
  supported_contracting_dims = ((act.ndim - 1,), (0,))
  # (lhs_contracting_dims, rhs_contracting_dims)
  if dimension_numbers != (
      supported_contracting_dims,
      ((), ()),
  ):  # (lhs_batch_dims, rhs_batch_dims)):
    raise ValueError(
        'Quantization is only supported for contracting dimension to be last '
        'dimension of lhs and first dimension of rhs.'
    )
  if quant_type == QuantType.AQT:
    # Let 's' be activation scales and 't' be weight scales. We implement
    # matmul(RoundAndClip(act*s), RoundAndClip(s^-1 * w * t)) *t^-1. In the
    # comments below, we refer to this terminology.
    # lax.dot accepts any combination of 1d and 2d arguments for its lhs and rhs
    # input. To simplify the AQT implementation, we only accept 2d arguments for
    # now.
    if w is not None:
      if act.shape[-1] != w.shape[0]:
        raise ValueError(
            'AQT is currently only implemented for matrix*matrix operations'
        )
      num_input_channels = act.shape[-1]

      if weight_params.axis is None:
        out_channel_shape = ()
      else:
        axes = _canonicalize_feature_axes(weight_params.axis, w.ndim)
        out_channel_shape = tuple(
            [w.shape[i] for i in range(w.ndim) if i not in axes]
        )

      # The ValueError raised in the guard at the beginning of this function
      # should have already checked that the weight matrix has a number of rows
      # equal to the number of channels in the activation.
      assert w.shape[0] == num_input_channels

      # We carry out all intermediate calculations using the same dtype as the
      # inputs. We want to be careful to not take a model configured to be
      # trained in bf16 and accidentally train it in fp32 by virtue of the scale
      # dtype being fp32.
      if act.dtype != w.dtype:
        raise TypeError(
            'Activations and weight must have the same dtype, but got'
            f' {act.dtype} and {w.dtype}'
        )
    else:
      assert quant_w is not None
    input_dtype = act.dtype

    is_act_quantized = False
    # In this case, activations will be quantized at some point during training
    # (either now or later) and so we need to gather activation statistics by
    # calling 'QuantOps.create_input_ops', even if activations are not being
    # quantized on this particular training step (see b/174516400).
    if act_hparams is not None and act_hparams.prec is not None:
      # Calculate 's', the per-column scale factor on activations.
      act_op = QuantOps.create_input_ops(
          act, hparams=act_hparams, bounds_params=bounds_params
      )
      is_act_quantized = act_op.should_quantize()
      # Quantize activation matrix by computing RoundAndClip(w*s)

      # TODO(malmaud): We have to cast quantized activations to an fp format
      # instead of int8 since int8 matmul with int32 accumulation is not yet
      # supported in XLA (and therefore in Jax). See b/170293520. We keep
      # 'act_quantized' in whatever it's original fp format was, typically bf16
      # or fp32, to follow what Fakequant does (see the type cast at the end of
      # QuantOpts.fake_quant).
      act_quantized = act_op.to_quantized(act, dtype=input_dtype)

      # Now calculate s^-1.  First we extract s, the activation scale factor,
      # into a  variable called 'act_scale'. We extract it from 'act_op', the
      # QuantOps instance that calculated the scale factors for the activation
      # matrix.
      act_scale = act_op.get_scale_for_aqt(allow_per_channel_scales=True)
      # act_scale should either be a scalar, corresponding to per-layer
      # quantization, or a matrix with shape (1,.., 1, num_input_channels),
      # corresponding to per-activation-channel scale factors.
      if act_scale.ndim != 0:
        shape_utils.assert_shapes_equal(
            act_scale.shape, (1,) * (act_scale.ndim - 1) + (num_input_channels,)
        )
        # 'w' has one row per column of 'act_scale'. To scale each row of 'w' by
        # the inverse of the corresponding column in 'act_scale', we first have
        # to reshape 'act_scale' from (1, ...,  num_input_channels) to
        # (num_input_channels, 1) so the scale factors will broadcast
        # appropriately across the columns of 'w'.
        act_scale = act_scale.reshape(num_input_channels, 1)
      # Now we calculate s^-1 * w.
      # TODO(shivaniagrawal): This section repeats code from the 'else' block.
      # The code is repeated twice because quantization can either be disabled
      # dynamically by setting the clipping bound to -1 (see comments on
      # 'should_quantize'), or statically by setting the 'prec' hyperparameter
      # to None. This block deals with the dynamic case (hence necessitating the
      # use of the dynamic 'lax.cond') while the 'else' block handles the static
      # case. Ultimately, we should unify them.
      act_quantized, act_scale = lax.cond(
          is_act_quantized,
          lambda _: (act_quantized, act_scale),
          lambda _: (  # pylint: disable=g-long-lambda
              act,
              jnp.ones_like(act_scale),
          ),
          None,
      )
    else:
      # In this case, activations are not being quantized; only weights. There
      # is no need to absorb activation scales into the rows of the weight
      # matrix so 'w_scaled_rows' can just be set to the original weight matrix.
      act_quantized = act
      act_scale = jnp.array(1.0, dtype=SCALE_DTYPE)

    is_weight_quantized = False
    if weight_params is not None and weight_params.prec is not None:
      is_weight_quantized = True
      if quant_w is not None:
        weight_quantized = quant_w.quantized_w.astype(input_dtype)
        weight_scale = quant_w.scale
      else:
        # Calculate 'r' from (s^-1) * w
        w_scaled_rows = ((1 / act_scale) * w).astype(input_dtype)
        weight_op = QuantOps.create_weights_ops(
            w_scaled_rows, weight_params=weight_params
        )
        weight_scale = weight_op.get_scale_for_aqt(
            allow_per_channel_scales=True
        )
        # Similar to 'act_scale' above, the weight_scale can either be a single
        # scalar or be a matrix with shape (1, out_channel_shape), corresponding
        # to a per-channel scale factor for the weight matrix. We verify it
        # here.
        if weight_scale.ndim != 0:
          shape_utils.assert_shapes_equal(
              weight_scale.shape, (1,) + out_channel_shape
          )
          if act.ndim != 0:
            weight_scale_shape = (1,) * (act.ndim - 1) + out_channel_shape
            weight_scale = weight_scale.reshape(weight_scale_shape)

        # Quantize weight matrix by calculating RoundAndClip(s^-1 * w * t)
        # TODO(malmaud): See comment on 'act_op.to_quantized' above, which
        # applies here as well.
        weight_quantized = weight_op.to_quantized(
            w_scaled_rows, dtype=input_dtype
        )
    else:
      assert (
          w is not None
      ), 'w can not be None if weight quantization is not specified.'
      weight_quantized = ((1 / act_scale) * w).astype(input_dtype)
      weight_scale = jnp.array(1.0, dtype=SCALE_DTYPE)

    # Use metadata context to annotate op metadata with quantization info
    act_prec = None if act_hparams is None else act_hparams.prec
    act_has_symm_distribution = act_hparams is not None and (
        act_hparams.input_distribution
        == QuantOps.ActHParams.InputDistribution.SYMMETRIC
    )
    weight_prec = None if weight_params is None else weight_params.prec

    # To decide whether to use an integer-domain dot operation, we first check
    # if the static quantization parameters are compatible with it by seeing if
    # they request that both inputs be quantized 8bits or less. Then check if
    # the dynamic parameters are compatible with it. ie, in a training run with
    # quantization enabled, are we past the activation start step yet.

    # We also do not use int8_to_int32_dot if activation has positive
    # distribution and prec=8, since we would not be able to fit uint8 range in
    # int8.
    # TODO(shivaniagrawal): A proper solution for this would be to have mixed
    # dot(uint8, int8) -> int32 in XLA.
    weight_fits_in_int8 = is_weight_quantized and (
        weight_prec is not None and weight_prec <= 8
    )
    # is_act_quantized might be an instance of a Jax tracer instead of a
    # Python boolean since it is generally computed from a dynamic input to a
    # JITted Jax function. Thus we use '&' instead of 'and'.
    act_prec_fits_int8 = act_prec is not None and (
        (act_prec == 8 and act_has_symm_distribution) or (act_prec < 8)
    )
    act_fits_in_int8 = is_act_quantized & act_prec_fits_int8
    use_int8_to_int32_dot = (
        prefer_int8_to_int32_dot & weight_fits_in_int8 & act_fits_in_int8
    )

    metadata_context = contextlib.suppress()
    if flags.FLAGS.metadata_enabled:
      metadata_context = compute_cost_utils.DotMetadataMonkeyPatch(
          lhs_prec=act_prec, rhs_prec=weight_prec, rhs_is_weight=True
      )

    with metadata_context:
      # Calculate matmul(...)
      out_quantized = dot_general_aqt(
          act_quantized,
          weight_quantized,
          dimension_numbers=dimension_numbers,
          dot_precision=dot_precision,
          use_int8_to_int32_dot=use_int8_to_int32_dot,
      )

    # Scale the columns of the matmul output by computing `matmul(...) * t^-1`
    # TODO(malmaud): Make it possible to return an unquantized matmul to support
    # disabling quantization during initial phase of training.
    #
    # We convert the return value back to input_dtype to ensure the output
    # tensor of quantized_dot has the same dtype as the input tensors to
    # quantized_dot. This explicit cast is necessary since if the inputs are
    # bf16, 'weight_scale' will still fp32 and so multipying out_quantized by
    # (1/weight_scale) will result in a fp32 tensor. We want to convert that
    # back to bf16.

    return (out_quantized * (1 / weight_scale)).astype(input_dtype)

  elif quant_type in (QuantType.FAKE_QUANT, QuantType.FAKE_QUANT_WITH_INT):
    if quant_w is not None:
      raise ValueError(
          'quantized vars are not supported in `fake_quant` style quantization.'
      )
    if quant_type == QuantType.FAKE_QUANT_WITH_INT:
      fake_dependency = act
    # create a dependency on fake input to control constant folding
    else:
      fake_dependency = None

    quantized_type = quant_type.to_jax_type()
    w = QuantOps.create_weights_fake_quant(
        w,
        weight_params=weight_params,
        quantized_type=quantized_type,
        fake_dependency=fake_dependency,
    )

    # TODO(shivaniagrawal): HParams currently allows act_hparams to be NONE.
    # Going forward we can change act_hparams to be required field where if
    # either `prec` or `bounds` is None will result in No activation
    # quantization.
    if act_hparams:
      act = QuantOps.create_inputs_fake_quant(
          act, hparams=act_hparams, bounds_params=bounds_params
      )

    metadata_context = contextlib.suppress()
    # Use metadata context to annotate op metadata with quantization info
    act_prec = None if act_hparams is None else act_hparams.prec
    weight_prec = None if weight_params is None else weight_params.prec

    if flags.FLAGS.metadata_enabled:
      metadata_context = compute_cost_utils.DotMetadataMonkeyPatch(
          lhs_prec=act_prec, rhs_prec=weight_prec, rhs_is_weight=True
      )
    with metadata_context:
      out_quantized = lax.dot_general(
          act, w, dimension_numbers=dimension_numbers, precision=dot_precision
      )
    return out_quantized
  else:
    raise RuntimeError(f'Unsupported quant_type {quant_type}')


class QuantizedDot(nn.Module):
  """Flax module that calculates a quantized 'dot' operation."""

  quant_type: QuantType
  weight_params: QuantOps.WeightParams
  act_hparams: Optional[QuantOps.ActHParams]
  prefer_int8_to_int32_dot: bool
  dot_precision: Optional[PrecisionType] = None

  # TODO(malmaud): Remove the 'padding_mask' field from 'GetBounds.Params'
  # so that 'bounds_params' can be a hyperparameter of this class and
  # only the padding mask will be passed as an argumen to '__call__'.
  @nn.compact
  def __call__(
      self,
      w: jnp.ndarray,
      act: jnp.ndarray,
      bounds_params: Union[
          None, get_bounds.GetBounds.Params, get_bounds.DynamicBounds.Params
      ],
  ) -> jnp.ndarray:
    return quantized_dot(
        w=w,
        act=act,
        bounds_params=bounds_params,
        quant_type=self.quant_type,
        weight_params=self.weight_params,
        act_hparams=self.act_hparams,
        dot_precision=self.dot_precision,
        prefer_int8_to_int32_dot=self.prefer_int8_to_int32_dot,
    )


def quantized_dynamic_dot_general(
    *,
    lhs_act: jnp.ndarray,
    rhs_act: jnp.ndarray,
    quant_type: QuantType,
    lhs_act_hparams: Optional[QuantOps.ActHParams],
    lhs_bounds_params: Union[
        None, get_bounds.GetBounds.Params, get_bounds.DynamicBounds.Params
    ],
    rhs_act_hparams: Optional[QuantOps.ActHParams],
    rhs_bounds_params: Union[
        None, get_bounds.GetBounds.Params, get_bounds.DynamicBounds.Params
    ],
    dot_dimension_numbers: lax.DotDimensionNumbers,
    dot_precision: Optional[PrecisionType] = None,
) -> jnp.ndarray:
  """LAX dot general with optionally quantized dynamic inputs.

  Wraps LAX's `DotGeneral
  <https://github.com/google/jax/blob/f65a327c764406db45e95048dfe09209d8ef6d37/jax/_src/lax/lax.py#L667`_
  operator.

  Args:
    lhs_act: an array representing weights
    rhs_act: an array representing activations
    quant_type: quantization strategy
    lhs_act_hparams: Optional activation quantization hyperparamers for lhs act;
      instance of QuantOps.ActHParams. None means no quantization.
    lhs_bounds_params: Optional get bounds params for lhs act auto quantization;
      instance of GetBounds.Params.
    rhs_act_hparams: Optional activation quantization hyperparamers for rhs act;
      instance of QuantOps.ActHParams. None means no quantization.
    rhs_bounds_params: Optional get bounds params for rhs act auto quantization;
      instance of GetBounds.Params.
    dot_dimension_numbers: a tuple of tuples of the form
      `((lhs_contracting_dims, rhs_contracting_dims), (lhs_batch_dims,
      rhs_batch_dims)).
    dot_precision: Optional. Either ``None``, which means the default precision
      for the backend, or a ``lax.Precision`` enum value (``Precision.DEFAULT``,
      ``Precision.HIGH`` or ``Precision.HIGHEST``).

  Returns:
    An array containing the result.

  Raises:
    RuntimeError: 'quant_type' had an unrecognized value.
    TypeError: Dtypes of lhs_act and rhs_act differed.
  """
  # See comment at the beginning of quantized_dot regarding its return type,
  # which also applies to this function.
  if quant_type == QuantType.AQT:
    # Let 's1' be the scale of 'lhs_act' and 's2' be the scale of 'rhs_act'.  We
    # calculate dot_general(RoundAndClip(s1*lhs_act),
    # RoundAndClip(s2*rhs_act))/(s1*s2). Note that unlike in
    # quantized_dot_general, the scale factors must be scalar (ie, per-tensor
    # quantization) since activations always have static scale factors and so
    # there is no way to absorb per-column scale factor from lhs_act into the
    # rows of rhs_act.

    # See comment on 'input_dtype' in 'quantized_dot'.
    if lhs_act.dtype != rhs_act.dtype:
      raise TypeError(
          'Both activations must have the same dtypes, but got '
          f'{lhs_act.dtype} and {rhs_act.dtype}'
      )
    input_dtype = lhs_act.dtype

    def get_tensor_and_scale_for_act(
        act: jnp.ndarray,
        hparams: Optional[QuantOps.ActHParams],
        bounds_params: Union[
            None, get_bounds.GetBounds.Params, get_bounds.DynamicBounds.Params
        ],
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
      # We check whether activations should be quantized based on 'hparams'. If
      # so, we quantize it. If not, we return it unchanged. In either case, we
      # return a scale factor appropriate for unscaling the result of the
      # lax.dot_general.
      if hparams is not None and hparams.prec is not None:
        quant_op = QuantOps.create_input_ops(
            act, hparams=hparams, bounds_params=bounds_params
        )

        scale = quant_op.get_scale_for_aqt(allow_per_channel_scales=False)
        # Since only per-layer scale factors are supported, we assert that the
        # scale factors are scalars.
        shape_utils.assert_shapes_compatible(scale.shape, ())
        # TODO(malmaud): See comment on 'act_op.to_quantized' earlier in this
        # file, which applies here as well.
        act_quantized = quant_op.to_quantized(act, dtype=input_dtype)

        # TODO(shivaniagrawal): See comment in 'dot_general' above on why this
        # logic is duplicated here and in the 'else' block below.
        return lax.cond(
            quant_op.should_quantize(),  #
            lambda _: (act_quantized, scale),  #
            lambda _: (act, jnp.array(1.0, dtype=SCALE_DTYPE)),  #
            None,
        )
      else:
        # To avoid having a separate code path for every possibility of which of
        # the two input tensors are quantized , we implement not quantizing an
        # activation tensor by simply setting its corresponding scale factor to
        # 1.0.
        return act, jnp.array(1.0, dtype=SCALE_DTYPE)

    lhs_quantized, lhs_scale = get_tensor_and_scale_for_act(
        lhs_act, lhs_act_hparams, lhs_bounds_params
    )
    rhs_quantized, rhs_scale = get_tensor_and_scale_for_act(
        rhs_act, rhs_act_hparams, rhs_bounds_params
    )

    metadata_context = contextlib.suppress()
    # Use metadata context to annotate op metadata with quantization info
    lhs_prec = None if lhs_act_hparams is None else lhs_act_hparams.prec
    rhs_prec = None if rhs_act_hparams is None else rhs_act_hparams.prec

    if flags.FLAGS.metadata_enabled:
      metadata_context = compute_cost_utils.DotMetadataMonkeyPatch(
          lhs_prec=lhs_prec, rhs_prec=rhs_prec, rhs_is_weight=False
      )
    with metadata_context:
      out_quantized = lax.dot_general(
          lhs_quantized,
          rhs_quantized,
          dimension_numbers=dot_dimension_numbers,
          precision=dot_precision,
      )

    # TODO(malmaud): There is an asymmetry here: when we scale the activations
    # to quantize them, the scaling happens in QuantOps.to_quantized. But here,
    # when we dequantize the matrix multiplication of the activations by
    # dividing by the product of the scale factors, we don't use QuantOps. It
    # would be cleaner to do both operations at the same level of abstraction.
    out = (out_quantized / (lhs_scale * rhs_scale)).astype(input_dtype)
  elif quant_type in (QuantType.FAKE_QUANT, QuantType.FAKE_QUANT_WITH_INT):
    # TODO(shivaniagrawal): HParams currently allows act_hparams to be NONE.
    # Going forward we can change act_hparams to be required field where if
    # either `prec` or `bounds` is None will result in No activation
    # quantization.
    if lhs_act_hparams:
      lhs_act = QuantOps.create_inputs_fake_quant(
          lhs_act, hparams=lhs_act_hparams, bounds_params=lhs_bounds_params
      )

    if rhs_act_hparams:
      rhs_act = QuantOps.create_inputs_fake_quant(
          rhs_act, hparams=rhs_act_hparams, bounds_params=rhs_bounds_params
      )

    metadata_context = contextlib.suppress()
    # Use metadata context to annotate op metadata with quantization info
    lhs_prec = None if lhs_act_hparams is None else lhs_act_hparams.prec
    rhs_prec = None if rhs_act_hparams is None else rhs_act_hparams.prec

    if flags.FLAGS.metadata_enabled:
      metadata_context = compute_cost_utils.DotMetadataMonkeyPatch(
          lhs_prec=lhs_prec, rhs_prec=rhs_prec, rhs_is_weight=False
      )
    with metadata_context:
      out = lax.dot_general(
          lhs_act,
          rhs_act,
          dimension_numbers=dot_dimension_numbers,
          precision=dot_precision,
      )
  else:
    raise RuntimeError(f'Unknown quant_type {quant_type}')
  return out


@functools.partial(jax.custom_jvp, nondiff_argnums=(1, 2, 3))
def quantized_sum(
    x: jnp.ndarray,  #
    axis: Union[int, Tuple[int, ...]],
    keepdims: bool,
    prec: Optional[QuantOps.FloatQuant.FloatPrec],
) -> jnp.ndarray:
  """Sums a tensor while quantizing intermediate accumulations.

  This is almost a drop-in replacement for jnp.sum. It only differs in that it
  takes in an 'act_hparams' parameter that controls the quantization of
  intermediate accumulations during the reduction.

  Arguments:
    x: Input, a Jax array
    axis: Which axes to reduce over (see jnp.sum docs)
    keepdims: Whether to keep of drop axes that are reduced (see jnp.sum docs)
    prec: Precision to quantize intermediate to. Currently can only an instance
      of QuantOps.FloatQuant.FloatPrec, corresponding to an unscaled
      floating-point format, or it can be None to indicate no quantization
      should be applied.

  Returns:
    A Jax array with the quantized sum of 'x'.
  """

  # Don't quantize. In this case, this function just wraps jnp.sum.
  if prec is None:
    return jnp.sum(x, axis=axis, keepdims=keepdims)

  # We bypass QuantOps.create_input_ops and directly call
  # QuantOps.create_symmetric_fp because the former creates an instance of
  # GetBounds, which in turn creates state variables to store activation
  # statistics. We do not want to compute statistics for each individual
  # addition within the sum reduction.
  fp_quant = QuantOps.FloatQuant(is_scaled=False, fp_spec=prec)
  quant_ops = QuantOps.create_symmetric_fp(fp_quant=fp_quant, bounds=None)

  if not isinstance(axis, Iterable):
    axis = (axis,)
  axis = utils.normalize_axes(axis, x.ndim)
  dtype = x.dtype

  zero = jnp.zeros((), dtype=dtype)
  x_quantized_sum = lax.reduce(
      x,
      init_values=zero,
      computation=lambda a, b: quant_ops.to_quantized(a + b, dtype=dtype),
      dimensions=axis,
  )

  if keepdims:
    x_quantized_sum = jnp.expand_dims(x_quantized_sum, axis)

  return x_quantized_sum


@quantized_sum.defjvp
def _quantized_sum_jvp(axis, keepdims, prec, primals, tangents):
  (x,), (x_dot,) = primals, tangents
  y = quantized_sum(x, axis=axis, keepdims=keepdims, prec=prec)
  # We calculate the JVP based on the JVP of the original jnp.sum function. That
  # corresponds to using a straight-through-estimator for the quantization
  # operators in 'quantized_sum'.
  _, y_dot = jax.jvp(
      lambda x: jnp.sum(x, keepdims=keepdims, axis=axis), (x,), (x_dot,)
  )
  return y, y_dot


@functools.partial(jax.custom_jvp, nondiff_argnums=(2, 3, 4))
def dot_general_aqt(
    lhs, rhs, dimension_numbers, dot_precision, use_int8_to_int32_dot
):
  """Wrapper around lax.dot_general, but with option to use integer dot.

  This function comes equipped with a custom gradient that defines the
  gradient of this function to be the same as the equivalent call to
  lax.dot_general, ignoring casts to and from integer types so that
  quantization-aware-training will work correctly.

  See docstring of lax.dot_general.

  Args:
    lhs: same as in lax.dot_general
    rhs: same as in lax.dot_general
    dimension_numbers: same as in lax.dot_general
    dot_precision: same as in lax.dot_general
    use_int8_to_int32_dot: boolean. If true, inputs to lax.dot_general will be
      cast to int8 and results accumulated to int32, then converted back to the
      original input type.

  Returns:
    Same as lax.dot_general.
  """

  # We define two versions of a dot operation. The first feeds lax.dot_general
  # the original inputs, which are typically bfloat16 or float32. The second
  # converts the inputs to int8 tensors and accumulates results to an int32
  # output.

  def dot_general_fp(ops):
    lhs_, rhs_ = ops
    return lax.dot_general(
        lhs_, rhs_, dimension_numbers=dimension_numbers, precision=dot_precision
    )

  def dot_general_int(ops):
    lhs_, rhs_ = ops
    input_dtype = lhs_.dtype
    lhs_int = lhs_.astype(jnp.int8)
    rhs_int = rhs_.astype(jnp.int8)
    return lax.dot_general(
        lhs_int,
        rhs_int,
        dimension_numbers=dimension_numbers,
        precision=dot_precision,
        preferred_element_type=jnp.int32,
    ).astype(input_dtype)

  return lax.cond(
      use_int8_to_int32_dot, dot_general_int, dot_general_fp, (lhs, rhs)
  )


@dot_general_aqt.defjvp
def _dot_general_aqt_jvp(
    dimension_numbers, dot_precision, use_int8_to_int32_dot, primals, tangents
):
  """Custom gradient for dot_general_aqt that ignores integer casts."""
  lhs, rhs = primals
  lhs_dot, rhs_dot = tangents
  y = dot_general_aqt(
      lhs,
      rhs,
      dimension_numbers=dimension_numbers,
      dot_precision=dot_precision,
      use_int8_to_int32_dot=use_int8_to_int32_dot,
  )

  def differentiable_dot_general(lhs_, rhs_):
    return lax.dot_general(
        lhs_, rhs_, dimension_numbers=dimension_numbers, precision=dot_precision
    )

  _, y_tangent = jax.jvp(
      differentiable_dot_general, (lhs, rhs), (lhs_dot, rhs_dot)
  )
  return y, y_tangent



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, Mapping, Sequence

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax.v2.flax import freezer
import flax
from flax import linen as nn
from flax.core import meta as nn_meta
import jax
from jax import numpy as jnp


@flax.struct.dataclass
class _CustomStructure:
  member: jnp.ndarray
  member_list: Sequence[jnp.ndarray]
  member_dict: Mapping[str, jnp.ndarray]


class TestModel(nn.Module):
  freezer_mode: freezer.FreezerMode
  axis_metadata_wrapper: None | Callable[..., nn_meta.AxisMetadata] = None

  def setup(self):
    self.f = freezer.Freezer(
        'freezer',
        mode=self.freezer_mode,
        axis_metadata_wrapper=self.axis_metadata_wrapper,
    )

  def __call__(self, x):
    """Emulates basic routine on how to use the freezer."""
    self.f.get()
    self.f.set(x)
    return x

  def freezer_get(self):
    return self.f.get()

  def freezer_set(self, x):
    return self.f.set(x)


class FreezerTest(parameterized.TestCase):
  def _create_custom_structure(self, prngkey):
    subkeys = jax.random.split(prngkey, 7)
    value1 = jax.random.normal(subkeys[0], (3, 3))
    value2 = [jax.random.normal(subkeys[i], (2, 2)) for i in [1, 2]]
    value3 = {'sub1': jax.random.normal(subkeys[4], (1, 4)),
              'sub2': jax.random.normal(subkeys[5], (3, 6)),
              'sub3': jax.random.normal(subkeys[6], (7, 7))}
    return _CustomStructure(value1, value2, value3)

  def _assert_same_tree_shape_dtype(self, tree1, tree2):
    """Checks if the two given pytrees have the same structure with the same leaves' shapes and dtypes."""
    leaves1, treedef1 = jax.tree.flatten(tree1)
    leaves2, treedef2 = jax.tree.flatten(tree2)

    # 1. Check if they have the same tree structure.
    self.assertEqual(treedef1, treedef2)

    # 2. Check if the leaves are with the same shapes and dtypes.
    self.assertEqual(len(leaves1), len(leaves2))
    for leaf1, leaf2 in zip(leaves1, leaves2):
      self.assertEqual(leaf1.shape, leaf2.shape)
      self.assertEqual(leaf1.dtype, leaf2.dtype)

  def test_freezer_get_set(self):
    class CustomWrapper(flax.struct.PyTreeNode, nn_meta.AxisMetadata):
      value: Any
      metadata: str = flax.struct.field(default=None, pytree_node=False)

      def unbox(self):
        return self.value

      def replace_boxed(self, val):
        return self.replace(value=val)

      def add_axis(self, index, params):
        return self

      def remove_axis(self, index, params):
        return self

    def axis_metadata_wrapper(x: _CustomStructure):
      ret = x.replace(
          member=CustomWrapper(x.member, 'member metadata'),
          member_list=[
              CustomWrapper(v, 'member list metadata') for v in x.member_list
          ],
      )
      return ret

    def unbox(x):
      if isinstance(x, CustomWrapper):
        return x.value
      return x

    subkeys = jax.random.split(jax.random.PRNGKey(0), 6)
    cs_for_init = self._create_custom_structure(subkeys[0])
    cs = self._create_custom_structure(subkeys[1])
    cs2 = self._create_custom_structure(subkeys[2])

    # 1. NONE mode test.
    tm_none = TestModel(freezer_mode=freezer.FreezerMode.NONE)
    param_init_none = tm_none.init(subkeys[3], cs_for_init)
    get_none = tm_none.apply(param_init_none, method=TestModel.freezer_get)
    _, param_set_none = tm_none.apply(
        param_init_none, cs, method=TestModel.freezer_set, mutable=True
    )
    get_after_set_none = tm_none.apply(
        param_set_none, method=TestModel.freezer_get
    )

    self.assertEqual(dict(), param_init_none)
    self.assertIsNone(get_none)
    self.assertIsNone(get_after_set_none)
    self.assertEqual(dict(), param_set_none)

    # 2. WRITE mode test.
    tm_write = TestModel(
        freezer_mode=freezer.FreezerMode.WRITE,
        axis_metadata_wrapper=axis_metadata_wrapper,
    )
    param_init_write = tm_write.init(subkeys[4], cs_for_init)

    # Check if the init parameters are properly wrapped.
    cs_frozen = param_init_write['freezer']['f']['frozen']
    self.assertIsInstance(cs_frozen.member, CustomWrapper)
    self.assertEqual(cs_frozen.member.metadata, 'member metadata')
    for value in cs_frozen.member_list:
      self.assertIsInstance(value, CustomWrapper)
      self.assertEqual(value.metadata, 'member list metadata')

    # Unbox the initialization parameters.
    # The metadata in the box is used to shard the variables here.
    param_init_write = jax.tree.map(
        unbox, param_init_write, is_leaf=lambda x: isinstance(x, CustomWrapper)
    )

    get_write = tm_write.apply(param_init_write, method=TestModel.freezer_get)
    _, param_set_write = tm_write.apply(
        param_init_write, cs, method=TestModel.freezer_set, mutable=True
    )
    get_after_set_write = tm_write.apply(
        param_set_write, method=TestModel.freezer_get
    )

    # Check if the init parameters are properly initialized with the init value.
    self._assert_same_tree_shape_dtype(
        cs_for_init, param_init_write['freezer']['f']['frozen']
    )

    # Since WRITE mode, get method should always return None.
    self.assertIsNone(get_write)
    self.assertIsNone(get_after_set_write)

    # Check if the parameter is properly updated after the set.
    self.assertEqual(cs, param_set_write['freezer']['f']['frozen'])

    # 3. READ mode test.
    tm_read = TestModel(
        freezer_mode=freezer.FreezerMode.READ,
        axis_metadata_wrapper=axis_metadata_wrapper,
    )
    param_init_read = tm_read.init(subkeys[5], cs_for_init)

    # Check if the init parameters are properly wrapped.
    cs_frozen = param_init_read['freezer']['f']['frozen']
    self.assertIsInstance(cs_frozen.member, CustomWrapper)
    self.assertEqual(cs_frozen.member.metadata, 'member metadata')
    for value in cs_frozen.member_list:
      self.assertIsInstance(value, CustomWrapper)
      self.assertEqual(value.metadata, 'member list metadata')

    # Unbox the initialization parameters.
    param_init_read = jax.tree.map(
        unbox, param_init_read, is_leaf=lambda x: isinstance(x, CustomWrapper)
    )

    # The tree structure initialized with READ should be the same with the one
    # after WRITE.
    self._assert_same_tree_shape_dtype(param_set_write, param_init_read)

    # get method should return the value specified by set.
    get_read = tm_read.apply(param_set_write, method=TestModel.freezer_get)

    self.assertEqual(cs, get_read)

    # set method should be ineffective.
    _, param_set_read = tm_read.apply(
        param_set_write, cs2, method=TestModel.freezer_set, mutable=True
    )
    get_after_set_read = tm_read.apply(
        param_set_read, method=TestModel.freezer_get
    )

    self.assertEqual(cs, get_after_set_read)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions to quantize models using AQT.

NOTE: All functions in this file assume lhs is an activation.
"""

import functools
from typing import Any

from aqt.jax.v2 import config as aqt_config
from aqt.jax.v2 import utils
from aqt.jax.v2.examples.cnn import model_utils
from aqt.jax.v2.flax import aqt_flax
from flax import linen as nn
from flax import typing as flax_typing
import jax
import jax.numpy as jnp

ModuleDef = Any


#########################################
# Common utility functions
#########################################
def _get_aqt_injected_model_kwargs(
    aqt_cfg: aqt_config.DotGeneral,
    lhs_quant_mode: utils.QuantMode,
    rhs_quant_mode: utils.QuantMode,
) -> dict[str, Any]:
  """Injects AQT functions to a given model class."""

  def _get_aqt_kwargs(
      aqt_cfg: aqt_config.DotGeneral,
      lhs_quant_mode: utils.QuantMode,
      rhs_quant_mode: utils.QuantMode,
  ) -> dict[str, Any]:
    """Gets AQT class init func prefilled with given configurations."""
    return dict(
        cfg=aqt_cfg,
        lhs_quant_mode=lhs_quant_mode,
        rhs_quant_mode=rhs_quant_mode,
        # For activations, we never freeze quantized values.
        lhs_freeze_mode=aqt_flax.FreezerMode.CALIBRATION,
        rhs_freeze_mode=aqt_flax.FreezerMode.CALIBRATION_AND_VALUE,
    )

  return dict(
      dot_general_cls=functools.partial(
          aqt_flax.AqtDotGeneral,
          **(_get_aqt_kwargs(aqt_cfg, lhs_quant_mode, rhs_quant_mode)),
      ),
      einsum_cls=functools.partial(
          aqt_flax.AqtEinsum,
          **(_get_aqt_kwargs(aqt_cfg, lhs_quant_mode, rhs_quant_mode)),
      ),
  )


#########################################
# Serving utility functions
#########################################
def _get_quantized_vars(
    model_cls: ModuleDef,
    aqt_cfg: aqt_config.DotGeneral,
    model_vars: flax_typing.FrozenVariableDict,
    act_calibrated: bool,
) -> flax_typing.FrozenVariableDict:
  """Gets quantized model variables."""
  # If activations have been calibrated, use the CONVERT mode to compute and
  # freeze scales from the statistics collected during the calibration pass.
  lhs_quant_mode = (
      utils.QuantMode.CONVERT if act_calibrated else utils.QuantMode.TRAIN
  )
  cnn_freeze = model_cls(
      bn_use_stats=False,
      **(
          _get_aqt_injected_model_kwargs(
              aqt_cfg, lhs_quant_mode, utils.QuantMode.CONVERT
          )
      ),
  )
  input_shape = (1, 28, 28, 1)
  _, quantized_param_vars = cnn_freeze.apply(
      model_vars,
      jnp.ones(input_shape),
      rngs={'params': jax.random.PRNGKey(0)},
      mutable=True,
  )
  return quantized_param_vars


def _get_serving_model(
    model_cls: ModuleDef, aqt_cfg: aqt_config.DotGeneral, act_calibrated: bool
) -> nn.Module:
  """Gets a Flax module to serve an AQT quantized model."""
  # If activations have been calibrated, use the SERVE mode to use the scales
  # frozen and stored in the model variables.
  lhs_quant_mode = (
      utils.QuantMode.SERVE if act_calibrated else utils.QuantMode.TRAIN
  )
  cnn_serve = model_cls(
      bn_use_stats=False,
      **(
          _get_aqt_injected_model_kwargs(
              aqt_cfg, lhs_quant_mode, utils.QuantMode.SERVE
          )
      ),
  )
  return cnn_serve


def serve_quantized(
    model_cls: ModuleDef,
    test_ds: dict[str, jax.Array],
    aqt_cfg: aqt_config.DotGeneral,
    model_vars: flax_typing.FrozenVariableDict,
    act_calibrated: bool,
) -> jax.Array:
  """Quantizes and tests a model with a given test dataset."""
  serve_model = _get_serving_model(
      model_cls, aqt_cfg, act_calibrated=act_calibrated
  )
  model_vars = _get_quantized_vars(
      model_cls, aqt_cfg, model_vars, act_calibrated=act_calibrated
  )
  return model_utils.serve(serve_model, model_vars, test_ds)


#########################################
# Calibration utility functions
#########################################
def _get_calibration_model(
    model_cls: ModuleDef, aqt_cfg: aqt_config.DotGeneral
) -> nn.Module:
  cnn_calibrate = model_cls(
      bn_use_stats=False,
      **(
          _get_aqt_injected_model_kwargs(
              aqt_cfg, utils.QuantMode.CALIBRATE, utils.QuantMode.CALIBRATE
          )
      ),
  )
  return cnn_calibrate


def _calibrate_epoch(
    calibration_model: nn.Module,
    model_vars: flax_typing.FrozenVariableDict,
    calibrate_ds: dict[str, jax.Array],
    batch_size: int,
    rng: jax.Array,
    calibration_steps: int,
) -> flax_typing.FrozenVariableDict:
  """Calibrates for a single epoch."""
  perms = model_utils.prepare_data_perm(
      calibrate_ds, batch_size, rng, calibration_steps
  )
  apply_fn = jax.jit(calibration_model.apply, static_argnames='mutable')
  for perm in perms:
    batch_images = calibrate_ds['image'][perm, ...]
    # Calibration simply updates model during inference. No need to compute loss
    # or gradients.
    _, model_vars = apply_fn(
        model_vars,
        batch_images,
        rngs={'params': jax.random.PRNGKey(0)},
        mutable=True,
    )
  return model_vars


def calibrate(
    model_cls: ModuleDef,
    aqt_cfg: aqt_config.DotGeneral,
    state: model_utils.TrainState,
    calibration_steps: int,
    calibrate_ds: dict[str, jax.Array],
) -> model_utils.TrainState:
  """Calibrates a given model."""
  rng = jax.random.key(0)
  batch_size = 128
  calibration_model = _get_calibration_model(model_cls, aqt_cfg)
  calibrated_vars = _calibrate_epoch(
      calibration_model,
      state.model_vars,
      calibrate_ds,
      batch_size,
      rng,
      calibration_steps,
  )

  return state.replace(model_vars=calibrated_vars)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Numerics for fp8."""

from typing import Literal, TypeAlias
from aqt.jax.v2 import utils
from aqt.jax.v2.numerics import numerics
import jax
import jax.numpy as jnp


FP8Dtype: TypeAlias = Literal['e4m3', 'e5m2']
fp8_map = {'e4m3': jnp.float8_e4m3fn, 'e5m2': jnp.float8_e5m2}


def fp_mantissa_round(x, mantissa_bits, key: jax.Array):
  """FP stochastic rounding for a given mantissa."""
  input_dtype = x.dtype
  total_bits = jnp.finfo(input_dtype).bits
  # (sign, exp, mantissa) = (1,8,x)
  msg = f'Unsupported dtype for sthochastic rounding: {input_dtype}.'
  assert input_dtype == jnp.bfloat16 or input_dtype == jnp.float32, msg
  # bf16 bit pattern: seeeeeeeemmmmmmm
  #                   7654321076543210
  mantissa = jnp.finfo(input_dtype).nmant
  bits_dtype = jnp.uint16 if total_bits == 16 else jnp.uint32
  assert mantissa_bits <= mantissa, 'bf16 mantissa bits'
  noise_bit_count = mantissa - mantissa_bits

  noise_mask = bits_dtype((1 << noise_bit_count) - 1)

  # TODO(lew): We can use smaller dtype
  x = jax.lax.bitcast_convert_type(x, bits_dtype)  # adds noise to mantissa
  noise = jax.random.bits(key, x.shape, bits_dtype)
  noise = jax.lax.convert_element_type(noise, bits_dtype) & noise_mask
  # This noise add might overflow up to sign bit if x(bf16) has max exponent.
  # In bf16 this happens if x=nan or x=inf.
  x = x + noise

  # TODO(lew): Is bitwise_and needed? Maybe round_to_nearest_even is ok.
  x = jax.lax.bitwise_and(x, ~noise_mask)  # zero-out lowest bits in mantissa
  x = jax.lax.bitcast_convert_type(x, input_dtype)

  return x


@utils.flax_slots_kw_only_dataclass
class Fp8Numerics(numerics.AqtNumerics):
  """Numerics for fp8."""

  # Storage type.
  dtype: Literal[jnp.float8_e4m3fn, jnp.float8_e5m2, jnp.bfloat16]

  # Requested rounding precision.
  exponent_bits: int = 4
  mantissa_bits: int = 3
  stochastic_rounding: bool = utils.static_field(default=False)

  def _get_edge_of_last_fp8_bucket(self):
    return jnp.finfo(self.dtype).max.astype(jnp.bfloat16)

  def get_dtype(self):
    return self.dtype

  def get_quant_bound(self):
    return self._get_edge_of_last_fp8_bucket()

  def vjp_fwd(self, x, context):
    match self.dtype:
      case jnp.float8_e4m3fn:
        assert (self.exponent_bits, self.mantissa_bits) == (4, 3)
      case jnp.float8_e5m2:
        assert (self.exponent_bits, self.mantissa_bits) == (5, 2)
      case jnp.bfloat16:
        assert self.exponent_bits <= 8
        assert self.mantissa_bits <= 7
      case _:
        assert False, f'Unsupported dtype: {self.dtype}'

    res = (x,)
    if not (
        (self.exponent_bits == 4 and self.mantissa_bits == 3)
        or (self.exponent_bits == 5 and self.mantissa_bits == 2)
    ):
      raise ValueError(
          '(exponent_bits, mantissa_bits) can only be (4,3) or (5,2) but was '
          f'({self.exponent_bits}, {self.mantissa_bits})'
      )

    if self.stochastic_rounding:
      msg = 'stochastic_rounding requires PRNGKey in Context'
      assert context.key is not None, msg
      x = fp_mantissa_round(x, self.mantissa_bits, context.key)

    # TODO(lew):
    #   - Is cliping good enough for exponent rounding?
    #   - Is it needed? Can we do clipping more efficiently?
    fwd_clip_bound = self._get_edge_of_last_fp8_bucket()
    x = jnp.clip(x, -1 * fwd_clip_bound, fwd_clip_bound)
    # TODO(lew): We can round more efficiently if stochastic_rounding == True.
    x = round_to_nearest_even(x, self.dtype)

    return x, res

  def vjp_bwd(self, res, grad):
    # This is gradient of clip.
    # For boundary values we will have full gradient.
    # We might use something like this for calibrations other than abs(max(x))
    # (x,) = res
    # ret = (x <= edge_of_last_bucket) * (x >= -edge_of_last_bucket) * grad
    del res
    ret = grad
    return (ret, None)


def round_to_nearest_even(x: jnp.ndarray, dtype: jnp.dtype) -> jnp.ndarray:
  x = x.astype(dtype)
  # TODO(lew): Is this rounding utilizing subnormals range fully?
  # bitcast_convert to uint8 to avoid allow_excess_precision set in XLA
  x = jax.lax.bitcast_convert_type(x, jnp.uint8)
  x = jax.lax.bitcast_convert_type(x, dtype)
  return x



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration dataclasses."""

# pylint: disable=g-importing-member
# pylint: disable=unused-import
# pylint: disable=g-explicit-bool-comparison

import copy
import functools
from typing import Literal, TypeAlias

from aqt.jax.v2 import aqt_dot_general
from aqt.jax.v2 import aqt_quantizer
from aqt.jax.v2 import calibration
from aqt.jax.v2 import stochastic_rounding
from aqt.jax.v2 import utils
# Temporary re-export from aqt.jax.v2.aqt_dot_general
# TODO(lew): Remove these imports, use setters instead
from aqt.jax.v2.aqt_conv_general import conv_general_dilated_make
from aqt.jax.v2.aqt_dot_general import CalibrationMode
from aqt.jax.v2.aqt_dot_general import DequantMode
from aqt.jax.v2.aqt_dot_general import dot_general_make
from aqt.jax.v2.aqt_dot_general import dot_general_raw_make
from aqt.jax.v2.aqt_dot_general import DotGeneral
from aqt.jax.v2.aqt_dot_general import DotGeneralRaw
from aqt.jax.v2.aqt_dot_general import dtypes_allowed_for_int32_accum
from aqt.jax.v2.aqt_dot_general import LocalAqt
from aqt.jax.v2.aqt_dot_general import Tensor
from aqt.jax.v2.aqt_quantizer import quantizer_make
from aqt.jax.v2.numerics import fp8_numerics
from aqt.jax.v2.numerics import fp_numerics
from aqt.jax.v2.numerics import int_numerics
from aqt.jax.v2.numerics import no_numerics
from aqt.jax.v2.numerics import numerics
from aqt.jax.v2.numerics import utils as numerics_utils
import jax
import jax.numpy as jnp


################################################################################
# Functions below are auxiliary config attribute setters.

# SKIP can be used as an argument to some set_xyz functions.
# It signals that set_xyz should make no changes to that option.
SKIP = 'skip'
SkipT: TypeAlias = Literal[SKIP]


def _split_key(key: None | jax.Array, num_splits: int):
  default = (None,) * num_splits
  return default if key is None else jax.random.split(key, num_splits)


def set_context(
    cfg: DotGeneral,
    key: None | jax.Array,
    train_step: None | int,
    lhs_quant_mode: utils.QuantMode = utils.QuantMode.TRAIN,
    rhs_quant_mode: utils.QuantMode = utils.QuantMode.TRAIN,
):
  """Set context with prng keys and train_steps for dot_general config."""

  def set_dg_raw_context(cfg_raw: DotGeneralRaw, key: None | jax.Array):
    key1, key2 = _split_key(key, num_splits=2)
    lhs_context = utils.Context(
        key=key1, train_step=train_step, quant_mode=lhs_quant_mode
    )
    rhs_context = utils.Context(
        key=key2, train_step=train_step, quant_mode=rhs_quant_mode
    )
    cfg_raw.dg_quantizer.set_context(lhs_context, rhs_context)

  key_fwd, key_dlhs, key_drhs = _split_key(key, num_splits=3)
  ret_cfg = copy.deepcopy(cfg)
  set_dg_raw_context(ret_cfg.fwd, key_fwd)
  set_dg_raw_context(ret_cfg.dlhs, key_dlhs)
  set_dg_raw_context(ret_cfg.drhs, key_drhs)
  return ret_cfg


def set_fwd_dequant_mode(
    cfg: DotGeneral,
    *,
    lhs_dequant_mode: None | DequantMode = None,
    rhs_dequant_mode: None | DequantMode = None,
):
  if lhs_dequant_mode is not None:
    cfg.fwd.lhs.dequant_mode = lhs_dequant_mode
  if rhs_dequant_mode is not None:
    cfg.fwd.rhs.dequant_mode = rhs_dequant_mode


def set_fwd_calibration_mode(
    cfg: DotGeneral,
    *,
    lhs_calibration_mode: CalibrationMode | SkipT = SKIP,
    rhs_calibration_mode: CalibrationMode | SkipT = SKIP,
):
  if lhs_calibration_mode != SKIP:
    cfg.fwd.lhs.calibration_mode = lhs_calibration_mode
  if rhs_calibration_mode != SKIP:
    cfg.fwd.rhs.calibration_mode = rhs_calibration_mode


def set_numerics(
    cfg: DotGeneralRaw,
    lhs_numerics: numerics.AqtNumerics,
    rhs_numerics: numerics.AqtNumerics,
):
  """Set numerics for DotGeneralRaw config."""
  assert isinstance(
      cfg.dg_quantizer, aqt_dot_general.DefaultDotGeneralQuantizer
  )
  cfg.dg_quantizer.lhs.numerics = lhs_numerics
  cfg.dg_quantizer.rhs.numerics = rhs_numerics
  if (
      lhs_numerics.get_dtype() in dtypes_allowed_for_int32_accum
      and rhs_numerics.get_dtype() in dtypes_allowed_for_int32_accum
  ):
    cfg.dg_accumulator_dtype = jnp.int32
  elif (
      lhs_numerics.get_dtype() in fp8_numerics.fp8_map.values()
      or rhs_numerics.get_dtype() in fp8_numerics.fp8_map.values()
  ):
    cfg.dg_accumulator_dtype = jnp.float32
  else:
    cfg.dg_accumulator_dtype = None


def set_fwd_rhs_dtype_int2(cfg: DotGeneral):
  """A special setter for int2 weights."""
  # Since XLA only supports int2 casting with an input shape of a multiple
  # of 128, we use this setter to enable int2 dtype.
  # Remove this setter and enable int2 in utils.infer_dtype_from_bits()
  # when XLA supports general int2 casting.
  assert isinstance(
      cfg.fwd.dg_quantizer.rhs.numerics, int_numerics.IntSymmetric
  )
  assert cfg.fwd.dg_quantizer.rhs.numerics.bits == 2
  # Disable pytype check since jnp.int2 is only dynamically to jax
  # when ml_dtypes package has it.
  cfg.fwd.dg_quantizer.rhs.numerics.dtype = jnp.int2  # pytype: disable=module-attr


def set_accumulator_dtype(
    cfg: DotGeneral,
    fwd_dtype: None | jnp.dtype | SkipT,
    dlhs_dtype: None | jnp.dtype | SkipT,
    drhs_dtype: None | jnp.dtype | SkipT,
):
  if fwd_dtype != SKIP:
    cfg.fwd.dg_accumulator_dtype = fwd_dtype
  if dlhs_dtype != SKIP:
    cfg.dlhs.dg_accumulator_dtype = dlhs_dtype
  if drhs_dtype != SKIP:
    cfg.drhs.dg_accumulator_dtype = drhs_dtype


def set_stochastic_rounding(
    cfg: DotGeneral,
    # Typically we have (but it's a caller's responsibility to check):
    # - vjp_lhs_stochastic_rounding is referring to the gradient and
    # - vjp_rhs_stochastic_rounding is referring to the activations/weights.
    vjp_lhs_stochastic_rounding: bool,
    vjp_rhs_stochastic_rounding: bool,
    implementation: str,
):
  """Configure stochastic rounding implementation."""
  noise_implementations = {
      'jax.uniform': stochastic_rounding.JaxUniform(),
      'custom-1': stochastic_rounding.RandomCenteredUniform(),
  }
  msg = f'{implementation} not supported.'
  assert implementation in noise_implementations.keys(), msg
  noise_fn = noise_implementations[implementation]
  assert isinstance(
      cfg.dlhs.dg_quantizer, aqt_dot_general.DefaultDotGeneralQuantizer
  )
  assert isinstance(
      cfg.drhs.dg_quantizer, aqt_dot_general.DefaultDotGeneralQuantizer
  )

  def _set_numerics_noise_fn(
      dg_quantizer: aqt_dot_general.DefaultDotGeneralQuantizer,
      is_lhs_sr: bool,
      is_rhs_sr: bool,
  ) -> None:
    # set stochastic noise for each dg_quantizer
    def _set_noise_fn(
        quantizer: aqt_quantizer.Quantizer,
        is_sr: bool,
    ) -> None:
      # set stochastic noise for each side of a dg_quantizer
      if isinstance(quantizer.numerics, fp8_numerics.Fp8Numerics):
        quantizer.numerics.stochastic_rounding = is_sr
      else:
        quantizer.numerics.noise_fn = noise_fn if is_sr else None

    _set_noise_fn(dg_quantizer.lhs, is_lhs_sr)
    _set_noise_fn(dg_quantizer.rhs, is_rhs_sr)

  is_lhs_sr = vjp_lhs_stochastic_rounding
  is_rhs_sr = vjp_rhs_stochastic_rounding
  _set_numerics_noise_fn(cfg.dlhs.dg_quantizer, is_lhs_sr, is_rhs_sr)
  _set_numerics_noise_fn(cfg.drhs.dg_quantizer, is_lhs_sr, is_rhs_sr)


def set_constant_calibration(
    cfg: DotGeneral, bound: float = 1.0, bias: None | float = None
):
  """Sets the static bound for calibration."""
  calibration_cls = functools.partial(
      calibration.ConstantCalibration, bound=bound, bias=bias
  )

  assert isinstance(
      cfg.fwd.dg_quantizer, aqt_dot_general.DefaultDotGeneralQuantizer
  )
  assert isinstance(
      cfg.dlhs.dg_quantizer, aqt_dot_general.DefaultDotGeneralQuantizer
  )
  assert isinstance(
      cfg.drhs.dg_quantizer, aqt_dot_general.DefaultDotGeneralQuantizer
  )

  cfg.fwd.dg_quantizer.lhs.calibration = calibration_cls
  cfg.fwd.dg_quantizer.rhs.calibration = calibration_cls
  cfg.dlhs.dg_quantizer.lhs.calibration = calibration_cls
  cfg.dlhs.dg_quantizer.rhs.calibration = calibration_cls
  cfg.drhs.dg_quantizer.lhs.calibration = calibration_cls
  cfg.drhs.dg_quantizer.rhs.calibration = calibration_cls


def set_local_aqt(
    cfg: DotGeneral,
    fwd_local_aqt: None | SkipT | LocalAqt,
    dlhs_local_aqt: None | SkipT | LocalAqt,
    drhs_local_aqt: None | SkipT | LocalAqt,
):
  if fwd_local_aqt != SKIP:
    cfg.fwd.local_aqt = fwd_local_aqt
  if dlhs_local_aqt != SKIP:
    cfg.dlhs.local_aqt = dlhs_local_aqt
  if drhs_local_aqt != SKIP:
    cfg.drhs.local_aqt = drhs_local_aqt


def set_use_fwd_quant(
    cfg: DotGeneral,
    dlhs_use_fwd_quant: None | bool | SkipT,
    drhs_use_fwd_quant: None | bool | SkipT,
):
  """Enable resusing of fwd pass quantization for backprop."""
  msg = 'use_fwd_quant is incompatible with use_mid_quant right now.'
  assert cfg.fwd.dg_quantizer.lhs_mid_alpha is None, msg
  assert cfg.fwd.dg_quantizer.rhs_mid_alpha is None, msg
  assert cfg.dlhs.dg_quantizer.lhs_mid_alpha is None, msg
  assert cfg.dlhs.dg_quantizer.rhs_mid_alpha is None, msg
  assert cfg.drhs.dg_quantizer.lhs_mid_alpha is None, msg
  assert cfg.drhs.dg_quantizer.rhs_mid_alpha is None, msg
  if dlhs_use_fwd_quant != SKIP:
    cfg.dlhs.rhs.use_fwd_quant = dlhs_use_fwd_quant
  if drhs_use_fwd_quant != SKIP:
    cfg.drhs.rhs.use_fwd_quant = drhs_use_fwd_quant


def set_use_mid_quant(
    cfg: DotGeneral,
    fwd_mid_alpha_both: SkipT | float,
    dlhs_mid_alpha_both: SkipT | float,
    drhs_mid_alpha_both: SkipT | float,
):
  """Enable middle quantization. Variant of SmoothQuant / AWQ."""
  assert isinstance(
      cfg.fwd.dg_quantizer, aqt_dot_general.DefaultDotGeneralQuantizer
  )
  assert isinstance(
      cfg.dlhs.dg_quantizer, aqt_dot_general.DefaultDotGeneralQuantizer
  )
  assert isinstance(
      cfg.drhs.dg_quantizer, aqt_dot_general.DefaultDotGeneralQuantizer
  )

  msg = 'use_fwd_quant is incompatible with use_mid_quant right now.'
  assert not cfg.dlhs.rhs.use_fwd_quant, msg
  assert not cfg.drhs.rhs.use_fwd_quant, msg

  @utils.flax_slots_kw_only_dataclass
  class DummyNumerics(numerics.AqtNumerics):
    """DummyNumerics for mid-quantization."""

    def get_quant_bound(self):
      return 1.0

    def get_dtype(self):
      assert False, 'Should not request dtype for mid-quantization.'

    def vjp_fwd(self, x, context):
      res = ()
      return x, res

    def vjp_bwd(self, res, grad):
      assert res == ()
      return grad

  calibration_cls = calibration.AbsMaxCalibration
  if fwd_mid_alpha_both != SKIP:
    cfg.fwd.dg_quantizer.lhs_mid.numerics = DummyNumerics()
    cfg.fwd.dg_quantizer.rhs_mid.numerics = DummyNumerics()
    cfg.fwd.dg_quantizer.lhs_mid.calibration = calibration_cls
    cfg.fwd.dg_quantizer.rhs_mid.calibration = calibration_cls
    cfg.fwd.dg_quantizer.lhs_mid_alpha = fwd_mid_alpha_both
    cfg.fwd.dg_quantizer.rhs_mid_alpha = fwd_mid_alpha_both
  if dlhs_mid_alpha_both != SKIP:
    cfg.dlhs.dg_quantizer.lhs_mid.numerics = DummyNumerics()
    cfg.dlhs.dg_quantizer.rhs_mid.numerics = DummyNumerics()
    cfg.dlhs.dg_quantizer.lhs_mid.calibration = calibration_cls
    cfg.dlhs.dg_quantizer.rhs_mid.calibration = calibration_cls
    cfg.dlhs.dg_quantizer.lhs_mid_alpha = dlhs_mid_alpha_both
    cfg.dlhs.dg_quantizer.rhs_mid_alpha = dlhs_mid_alpha_both
  if drhs_mid_alpha_both != SKIP:
    cfg.drhs.dg_quantizer.lhs_mid.numerics = DummyNumerics()
    cfg.drhs.dg_quantizer.rhs_mid.numerics = DummyNumerics()
    cfg.drhs.dg_quantizer.lhs_mid.calibration = calibration_cls
    cfg.drhs.dg_quantizer.rhs_mid.calibration = calibration_cls
    cfg.drhs.dg_quantizer.lhs_mid_alpha = drhs_mid_alpha_both
    cfg.drhs.dg_quantizer.rhs_mid_alpha = drhs_mid_alpha_both


def set_int_numerics_preserve_zero(cfg: DotGeneral, preserve_zero: bool):
  """Set preserve_zero for int_numerics."""
  assert isinstance(
      cfg.fwd.dg_quantizer, aqt_dot_general.DefaultDotGeneralQuantizer
  )
  assert isinstance(
      cfg.dlhs.dg_quantizer, aqt_dot_general.DefaultDotGeneralQuantizer
  )
  assert isinstance(
      cfg.drhs.dg_quantizer, aqt_dot_general.DefaultDotGeneralQuantizer
  )

  for dot_general_raw in [cfg.fwd, cfg.dlhs, cfg.drhs]:
    dg_quantizer = dot_general_raw.dg_quantizer
    for q_numerics in [dg_quantizer.lhs.numerics, dg_quantizer.rhs.numerics]:
      if isinstance(q_numerics, int_numerics.IntSymmetric):
        q_numerics.preserve_zero = preserve_zero
        updated_dtype = (
            utils.infer_dtype_from_bits(q_numerics.bits)
            if preserve_zero
            else None
        )
        q_numerics.dtype = updated_dtype


def set_auto_calib_scale(
    cfg: DotGeneral, auto_clip_search_config: utils.AutoScaleSearchConfig
) -> None:
  """Update `cfg`'s quantizers' calibration to use auto clipping search.

  Currently only supports the weights (rhs) of `DotGeneral`, since the iterative
  process of finding the scale tensors might be too slow for the activations
  (lhs).

  Args:
    cfg: The config to be updated.
    auto_clip_search_config: The config for auto clipping search.
  """
  assert isinstance(
      cfg.fwd.dg_quantizer, aqt_dot_general.DefaultDotGeneralQuantizer
  )
  assert isinstance(
      cfg.dlhs.dg_quantizer, aqt_dot_general.DefaultDotGeneralQuantizer
  )
  assert isinstance(
      cfg.drhs.dg_quantizer, aqt_dot_general.DefaultDotGeneralQuantizer
  )

  for dot_general_raw in [cfg.fwd, cfg.dlhs, cfg.drhs]:
    quantizer = dot_general_raw.dg_quantizer.rhs
    # TODO(lew): Remove partial inspection wherever possible.
    # Partial inspection is needed because the current implementation of delayed
    # calibration initialization requires the configuration to be set via
    # functools.partial.
    keywords = {}
    if isinstance(quantizer.calibration, functools.partial):
      keywords = quantizer.calibration.keywords
    keywords.update(auto_clip_search_config=auto_clip_search_config)
    quantizer.calibration = functools.partial(
        calibration.SnrBasedAutoCalibration, **keywords
    )


def set_absmax_calib_scale(cfg: DotGeneral, scale: float):
  """Set clipping_scale and clip_gradient for AbsMaxCalibration quantizers.

  Does not modify the configuration for quantizers with other calibration
  classes or None calibration.

  Args:
    cfg: The config to be updated.
    scale: The clipping scale.
  """
  assert isinstance(
      cfg.fwd.dg_quantizer, aqt_dot_general.DefaultDotGeneralQuantizer
  )
  assert isinstance(
      cfg.dlhs.dg_quantizer, aqt_dot_general.DefaultDotGeneralQuantizer
  )
  assert isinstance(
      cfg.drhs.dg_quantizer, aqt_dot_general.DefaultDotGeneralQuantizer
  )

  for dot_general_raw in [cfg.fwd, cfg.dlhs, cfg.drhs]:
    dg_quantizer = dot_general_raw.dg_quantizer
    for quantizer in [dg_quantizer.lhs, dg_quantizer.rhs]:
      calibration_cls = quantizer.calibration
      if calibration_cls is None:
        continue

      # TODO(lew): Remove partial inspection wherever possible.
      # Partial inspection is needed because the current implementation of
      # delayed calibration initialization requires the configuration to be set
      # via functools.partial.
      keywords = {}
      if isinstance(calibration_cls, functools.partial):
        keywords = calibration_cls.keywords
        calibration_cls = calibration_cls.func
      keywords.update(clipping_scale=scale)
      assert calibration_cls == calibration.AbsMaxCalibration, (
          'clipping_scale is only available in AbsMaxCalibration, while'
          f' {quantizer.calibration} is used in current config.'
      )
      quantizer.calibration = functools.partial(
          calibration.AbsMaxCalibration,
          **keywords,
      )
      if scale < 1.0 and isinstance(
          quantizer.numerics, int_numerics.IntSymmetric
      ):
        quantizer.numerics.clip_gradient = True


def set_bits(
    cfg: DotGeneral,
    fwd_lhs_bit: None | int | fp8_numerics.FP8Dtype,
    fwd_rhs_bit: None | int | fp8_numerics.FP8Dtype,
    dlhs_lhs_bit: None | int | fp8_numerics.FP8Dtype,
    dlhs_rhs_bit: None | int | fp8_numerics.FP8Dtype,
    drhs_lhs_bit: None | int | fp8_numerics.FP8Dtype,
    drhs_rhs_bit: None | int | fp8_numerics.FP8Dtype,
) -> DotGeneral:
  """Set quant bits for dot_general. Overwrites with AbsMaxCalibration."""
  calibration_cls = calibration.AbsMaxCalibration

  set_numerics(
      cfg.fwd,
      numerics_utils.get_numerics(fwd_lhs_bit),
      numerics_utils.get_numerics(fwd_rhs_bit),
  )
  if fwd_lhs_bit is not None:
    cfg.fwd.dg_quantizer.lhs.calibration = calibration_cls
  if fwd_rhs_bit is not None:
    cfg.fwd.dg_quantizer.rhs.calibration = calibration_cls

  set_numerics(
      cfg.dlhs,
      numerics_utils.get_numerics(dlhs_lhs_bit),
      numerics_utils.get_numerics(dlhs_rhs_bit),
  )
  if dlhs_lhs_bit is not None:
    cfg.dlhs.dg_quantizer.lhs.calibration = calibration_cls
  if dlhs_rhs_bit is not None:
    cfg.dlhs.dg_quantizer.rhs.calibration = calibration_cls

  set_numerics(
      cfg.drhs,
      numerics_utils.get_numerics(drhs_lhs_bit),
      numerics_utils.get_numerics(drhs_rhs_bit),
  )
  if drhs_lhs_bit is not None:
    cfg.drhs.dg_quantizer.lhs.calibration = calibration_cls
  if drhs_rhs_bit is not None:
    cfg.drhs.dg_quantizer.rhs.calibration = calibration_cls

  # use_fwd_quant is by default set to False if fwd pass is quantized.
  # This is to make the configuration logically correct,
  # i.e., use_fwd_quant cannot be None when fwd is quantized.
  # It is user's responsibility to further choose between False and True.
  dlhs_use_fwd_quant = False if fwd_rhs_bit is not None else SKIP
  drhs_use_fwd_quant = False if fwd_lhs_bit is not None else SKIP
  set_use_fwd_quant(cfg, dlhs_use_fwd_quant, drhs_use_fwd_quant)
  return cfg


def set_scale_and_bias_dtype(cfg: DotGeneral, dtype: jnp.dtype):
  """Set the dtype for all scales and biases in the given DotGeneral config."""
  assert isinstance(
      cfg.fwd.dg_quantizer, aqt_dot_general.DefaultDotGeneralQuantizer
  )
  assert isinstance(
      cfg.dlhs.dg_quantizer, aqt_dot_general.DefaultDotGeneralQuantizer
  )
  assert isinstance(
      cfg.drhs.dg_quantizer, aqt_dot_general.DefaultDotGeneralQuantizer
  )

  def _update_dtype(quantizer: aqt_quantizer.Quantizer):
    calibration_cls = quantizer.calibration
    if calibration_cls is None:
      return

    # TODO(lew): Remove partial inspection wherever possible.
    # Partial inspection is needed because the current implementation of delayed
    # calibration initialization requires the configuration to be set via
    # functools.partial.
    keywords = {}
    if isinstance(calibration_cls, functools.partial):
      keywords = calibration_cls.keywords
      calibration_cls = calibration_cls.func
    keywords.update(dtype=dtype)
    quantizer.calibration = functools.partial(calibration_cls, **keywords)

  _update_dtype(cfg.fwd.dg_quantizer.lhs)
  _update_dtype(cfg.fwd.dg_quantizer.rhs)
  _update_dtype(cfg.dlhs.dg_quantizer.lhs)
  _update_dtype(cfg.dlhs.dg_quantizer.rhs)
  _update_dtype(cfg.drhs.dg_quantizer.lhs)
  _update_dtype(cfg.drhs.dg_quantizer.rhs)


################################################################################
# Functions below are auxiliary config creators.


def default_unquantized_config() -> DotGeneral:
  """Aqt config for floating-point dot general."""

  def tensor_cfg() -> Tensor:
    cfg = Tensor(
        use_fwd_quant=False,
        dequant_mode=DequantMode.OUTPUT,
        calibration_mode=CalibrationMode.CONTRACTING_AXIS,
    )
    return cfg

  def quantizer() -> aqt_quantizer.Quantizer:
    return aqt_quantizer.Quantizer(
        numerics=no_numerics.NoNumerics(),
        calib_shared_axes=None,
        scale_stop_grad=True,
        calibration=None,
        context=utils.Context(key=None, train_step=None),
    )

  def dg_raw_cfg(jax_scope_name: str) -> DotGeneralRaw:
    return DotGeneralRaw(
        lhs=tensor_cfg(),
        rhs=tensor_cfg(),
        dg_quantizer=aqt_dot_general.DefaultDotGeneralQuantizer(
            lhs=quantizer(),
            rhs=quantizer(),
            lhs_mid=quantizer(),
            rhs_mid=quantizer(),
        ),
        dg_accumulator_dtype=None,
        local_aqt=None,
        jax_scope_name=jax_scope_name,
    )

  dg_cfg = DotGeneral(
      fwd=dg_raw_cfg('aqt_fwd'),
      dlhs=dg_raw_cfg('aqt_dlhs'),
      drhs=dg_raw_cfg('aqt_drhs'),
  )
  return dg_cfg


def fully_quantized(
    *,
    fwd_bits: None | int = 8,
    bwd_bits: None | int = 8,
    use_fwd_quant: bool = True,
    use_stochastic_rounding: None | bool = True,
    # Typically we have (but it's a caller's responsibility to check):
    # - vjp_lhs_stochastic_rounding is referring to the gradient and
    # - vjp_rhs_stochastic_rounding is referring to the activations/weights.
    vjp_lhs_stochastic_rounding: None | bool = None,
    vjp_rhs_stochastic_rounding: None | bool = None,
    # The dummy static bound flag is temporary, for performance benchmarking.
    use_dummy_static_bound: bool = False,
    dlhs_local_aqt: None | LocalAqt = None,
    drhs_local_aqt: None | LocalAqt = None,
) -> DotGeneral:
  """Fully Quantized Training."""
  cfg = dot_general_make(
      lhs_bits=fwd_bits,
      rhs_bits=fwd_bits,
      bwd_bits=bwd_bits,
      use_fwd_quant=use_fwd_quant,
      dlhs_local_aqt=dlhs_local_aqt,
      drhs_local_aqt=drhs_local_aqt,
  )

  # Stochastic Rounding
  # These 3 variables are used to ensure we don't mix
  # old and new style of SR configuration.
  old_style_sr_config = use_stochastic_rounding is not None
  new_style_sr_config_lhs = vjp_lhs_stochastic_rounding is not None
  new_style_sr_config_rhs = vjp_rhs_stochastic_rounding is not None
  assert new_style_sr_config_lhs == new_style_sr_config_rhs, (
      'if you use new style SR config (vjp_xhs_stochastic_rounding), do pass'
      ' both lhs and rhs explicitely.'
  )
  assert new_style_sr_config_lhs != old_style_sr_config

  true = True  # A crude way to get around g-explicit-bool-comparison warning

  assert not (vjp_lhs_stochastic_rounding and vjp_rhs_stochastic_rounding), (
      'This config is buggy when you set both to True. Contact lew@ or use'
      ' config_v3'
  )

  # By default use jax.uniform for stochastic rounding
  if use_stochastic_rounding == true:
    set_stochastic_rounding(cfg, True, True, 'jax.uniform')

  if vjp_lhs_stochastic_rounding == true:
    set_stochastic_rounding(cfg, True, False, 'jax.uniform')

  if vjp_rhs_stochastic_rounding == true:
    set_stochastic_rounding(cfg, False, True, 'jax.uniform')

  if use_dummy_static_bound:
    set_constant_calibration(cfg, 1.0)

  assert cfg.fwd.local_aqt is None, 'local_aqt is not yet supported in fwd.'

  return cfg


def config_v3(
    *,
    fwd_bits: None | int = 8,
    dlhs_bits: None | int = 8,
    drhs_bits: None | int = None,
    # The dummy static bound flag is for performance benchmarking.
    use_dummy_static_bound: bool = False,
    rng_type: str = 'jax.uniform',  # 'custom-1'
    dlhs_local_aqt: None | LocalAqt = None,
    drhs_local_aqt: None | LocalAqt = None,
    fwd_accumulator_dtype: ... = jnp.int32,
    dlhs_accumulator_dtype: ... = jnp.int32,
    drhs_accumulator_dtype: ... = None,
) -> DotGeneral:
  """Fully Quantized Training."""
  fwd = dot_general_raw_make(
      fwd_bits,
      fwd_bits,
      jax_scope_name='aqt_fwd',
      initialize_calibration=False,
  )
  dlhs = dot_general_raw_make(
      dlhs_bits,
      dlhs_bits,
      local_aqt=dlhs_local_aqt,
      jax_scope_name='aqt_dlhs',
      initialize_calibration=False,
  )
  drhs = dot_general_raw_make(
      drhs_bits,
      drhs_bits,
      local_aqt=drhs_local_aqt,
      jax_scope_name='aqt_drhs',
      initialize_calibration=False,
  )
  cfg = DotGeneral(fwd=fwd, dlhs=dlhs, drhs=drhs)

  cfg.dlhs.rhs.use_fwd_quant = False
  cfg.drhs.rhs.use_fwd_quant = False

  # Typically we have (but I don't know if it is guaranteed):
  # - vjp_lhs_stochastic_rounding is referring to the gradient and
  # - vjp_rhs_stochastic_rounding is referring to the activations/weights.
  set_stochastic_rounding(
      cfg,
      vjp_lhs_stochastic_rounding=True,
      vjp_rhs_stochastic_rounding=False,
      implementation=rng_type,
  )

  if use_dummy_static_bound:
    set_constant_calibration(cfg, 1.0)

  set_accumulator_dtype(
      cfg,
      fwd_dtype=fwd_accumulator_dtype,
      dlhs_dtype=dlhs_accumulator_dtype,
      drhs_dtype=drhs_accumulator_dtype,
  )
  assert cfg.fwd.local_aqt is None, 'local_aqt is not yet supported in fwd.'
  return cfg


def config_v4(
    *,
    fwd_bits: None | int | fp8_numerics.FP8Dtype = 8,
    dlhs_bits: None | int | fp8_numerics.FP8Dtype = 8,
    drhs_bits: None | int | fp8_numerics.FP8Dtype = None,
    # The dummy static bound flag is for performance benchmarking.
    use_dummy_static_bound: bool = False,
    rng_type: str = 'jax.uniform',  # 'custom-1'
    dlhs_local_aqt: None | LocalAqt = None,
    drhs_local_aqt: None | LocalAqt = None,
    # accumulator dtype by default is automatically set in set_bits,
    # but users can still configure a special dtype such as jnp.int16, etc.
    fwd_accumulator_dtype: None | jnp.dtype | SkipT = SKIP,
    dlhs_accumulator_dtype: None | jnp.dtype | SkipT = SKIP,
    drhs_accumulator_dtype: None | jnp.dtype | SkipT = SKIP,
    dlhs_use_fwd_quant: None | bool | SkipT = SKIP,
    drhs_use_fwd_quant: None | bool | SkipT = SKIP,
    fwd_mid_alpha_both: SkipT | float = SKIP,
    dlhs_mid_alpha_both: SkipT | float = SKIP,
    drhs_mid_alpha_both: SkipT | float = SKIP,
) -> DotGeneral:
  """Version 4 of user-visible AQT config."""
  cfg = default_unquantized_config()
  set_bits(
      cfg,
      fwd_lhs_bit=fwd_bits,
      fwd_rhs_bit=fwd_bits,
      dlhs_lhs_bit=dlhs_bits,
      dlhs_rhs_bit=dlhs_bits,
      drhs_lhs_bit=drhs_bits,
      drhs_rhs_bit=drhs_bits,
  )
  set_accumulator_dtype(
      cfg,
      fwd_dtype=fwd_accumulator_dtype,
      dlhs_dtype=dlhs_accumulator_dtype,
      drhs_dtype=drhs_accumulator_dtype,
  )
  set_stochastic_rounding(
      cfg,
      vjp_lhs_stochastic_rounding=True,
      vjp_rhs_stochastic_rounding=False,
      implementation=rng_type,
  )
  if use_dummy_static_bound:
    set_constant_calibration(cfg, 1.0)
  set_local_aqt(
      cfg,
      fwd_local_aqt=SKIP,
      dlhs_local_aqt=dlhs_local_aqt,
      drhs_local_aqt=drhs_local_aqt,
  )
  set_use_fwd_quant(
      cfg,
      dlhs_use_fwd_quant=dlhs_use_fwd_quant,
      drhs_use_fwd_quant=drhs_use_fwd_quant,
  )
  set_use_mid_quant(
      cfg,
      fwd_mid_alpha_both=fwd_mid_alpha_both,
      dlhs_mid_alpha_both=dlhs_mid_alpha_both,
      drhs_mid_alpha_both=drhs_mid_alpha_both,
  )
  assert cfg.fwd.local_aqt is None, 'local_aqt is not yet supported in fwd.'
  return cfg


def config_fwd_fp8(fwd_bits: fp8_numerics.FP8Dtype = 'e4m3') -> DotGeneral:
  """Configs for FP8 forward pass."""
  assert (
      fwd_bits in fp8_numerics.fp8_map.keys()
  ), 'FP8 only supports 4 or 5 exponent bits'
  cfg = config_v4(fwd_bits=8, dlhs_bits=None, drhs_bits=None)
  set_bits(
      cfg,
      fwd_lhs_bit=fwd_bits,
      fwd_rhs_bit=fwd_bits,
      dlhs_lhs_bit=None,
      dlhs_rhs_bit=None,
      drhs_lhs_bit=None,
      drhs_rhs_bit=None,
  )
  set_stochastic_rounding(cfg, False, False, 'jax.uniform')
  assert cfg.fwd.local_aqt is None, 'local_aqt is not yet supported in fwd.'
  return cfg


def set_fwd_calibration(cfg: DotGeneral, calibration_factory) -> DotGeneral:
  """Updates aqt_cfg for static range calibration."""
  assert isinstance(
      cfg.fwd.dg_quantizer, aqt_dot_general.DefaultDotGeneralQuantizer
  )

  cfg.fwd.dg_quantizer.lhs.calibration = calibration_factory
  cfg.fwd.dg_quantizer.rhs.calibration = calibration_factory



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper class and methods for computing and tracking statistics."""

import functools
from typing import Any, Callable, Iterable, Optional
from aqt.jax_legacy.jax import shape_utils
from aqt.jax_legacy.jax.flax import struct as flax_struct
import flax
import jax
from jax import lax
import jax.numpy as jnp


def is_in_pmapped_context(paxis_name: Optional[str]) -> bool:
  """Returns whether context is pmapped."""
  return paxis_name is not None and len(jax.devices()) > 1


def masked_reduction(
    x: jnp.ndarray,
    *,
    mask: Optional[jnp.ndarray],
    axis: Iterable[int],
    paxis_name: Optional[str],
    local_reduction_fn: Callable[..., jnp.ndarray],
    pmap_reduction_fn: Callable[..., jnp.ndarray],
    identity: float,
    keepdims: bool,
) -> jnp.ndarray:
  """Reduce an array locally and across devices, excluding masked elements.

  Args:
    x: Tensor to take the mean of.
    mask: Boolean array of same shape as 'x'. True elements are included in the
      mean, false elements are excluded. None means to not mask out any elements
    axis: Axis of 'x' to compute the mean over.
    paxis_name: Optional. If not None, will take a distributed mean of 'x'
      across devices using the specified parallel axis.
    local_reduction_fn: Reduction function to apply locally to 'x'.
    pmap_reduction_fn: A distributed reduction function to apply after 'x' has
      been reduced locally. A member of the lax.p* family of functions.
    identity: Identity element for the reduction performs by 'reduction_fn'. eg,
      0 for sum, 1 for multipy, etc.
    keepdims: Same meaning as the corresponding parameter in `numpy.mean`.
      Whether to keep the reduction axes or squeeze them out.

  Returns:
    A Jax array resulting from applying the reduction to 'x'.
  """
  if mask is None:
    x_masked = x
  else:
    if x.shape != mask.shape:
      raise ValueError(
          'Masked reduction received input and mask with different shapes')
    if mask.dtype != jnp.bool_:
      raise ValueError(
          f'Mask should be boolean-typed, but has dtype {mask.dtype}')

    x_masked = jnp.where(mask, x, identity)
  x_reduced = local_reduction_fn(x_masked, axis=axis, keepdims=keepdims)
  if is_in_pmapped_context(paxis_name):
    x_reduced = pmap_reduction_fn(x_reduced, axis_name=paxis_name)
  return x_reduced


masked_sum = functools.partial(
    masked_reduction,
    local_reduction_fn=jnp.sum,
    pmap_reduction_fn=lax.psum,
    identity=0.0)

# TODO(shivaniagrawal): The pmap_function_fn used here should be 'lax.pmax' for
# masked_max and 'lax.pmin' for masked_min, but Jax currently doesn't support
# using those in multihost training. Jax team says it will be trivial to add and
# they welcome PRs, so we just approximate distributed max using lax.pmean until
# we get that PR into Jax.
masked_mean_of_max = functools.partial(
    masked_reduction,
    local_reduction_fn=jnp.max,
    pmap_reduction_fn=lax.pmean,
    identity=-jnp.inf)
masked_mean_of_min = functools.partial(
    masked_reduction,
    local_reduction_fn=jnp.min,
    pmap_reduction_fn=lax.pmean,
    identity=jnp.inf)


def masked_mean(x: jnp.ndarray, *, mask: jnp.ndarray, axis: Iterable[int],
                paxis_name: Optional[str], keepdims: bool) -> jnp.ndarray:
  """Calculates the mean of a tensor, excluding masked-out entries.

  Args:
    x: Tensor to take the mean of.
    mask: Boolean array of same shape as 'x'. True elements are included in the
      mean, false elements are excluded.
    axis: Axis of 'x' to compute the mean over.
    paxis_name: Optional. If not None, will take a distributed mean of 'x'
      across devices using the specified parallel axis.
    keepdims: Same meaning as the corresponding parameter in `numpy.mean`.
      Whether to keep the reduction axes or squeeze them out.

  Returns:
    Tensor resulting from reducing 'x' over axes in 'axis'.
  """
  assert x.shape == mask.shape
  x_masked_sum = masked_sum(
      x, mask=mask, axis=axis, paxis_name=paxis_name, keepdims=keepdims)
  mask_count = masked_sum(
      x=mask, mask=None, axis=axis, paxis_name=paxis_name, keepdims=keepdims)
  x_masked_mean = x_masked_sum / mask_count
  return x_masked_mean


@flax_struct.dataclass
class Stats:
  """Dataclass to keep track of statistics."""

  n: int  # samples count
  mean: jnp.ndarray  # mean of values over axis
  mean_abs: jnp.ndarray  # mean of absolute values over axis
  mean_sq: jnp.ndarray  # mean of square values over axis
  # minimum value over axis in each device's batch, averaged over devices.
  mean_batch_minimum: jnp.ndarray
  # maximum value over axis in each device's batch, averaged over devices.
  mean_batch_maximum: jnp.ndarray

  @classmethod
  def stats_initializer(cls, shape: Iterable[int], *, dtype: Any = jnp.float32):
    """Constructor to init a empty Stats instance.

    Args:
      shape: shape of the statistics (mean, mean_sq and mean_abs)
      dtype: the dtype of the stats (default: float32).

    Returns:
      A new instance of Stats, with statistics intialized to 0.
    """
    key = jax.random.PRNGKey(
        1)  # no effect, but needed for flax.linen.initializers.zeros.
    shape = tuple(shape)
    return cls(
        n=0,
        mean=flax.linen.initializers.zeros(key, shape, dtype),
        mean_abs=flax.linen.initializers.zeros(key, shape, dtype),
        mean_sq=flax.linen.initializers.zeros(key, shape, dtype),
        mean_batch_maximum=flax.linen.initializers.zeros(key, shape, dtype),
        mean_batch_minimum=flax.linen.initializers.zeros(key, shape, dtype))  # pytype: disable=wrong-keyword-args  # trace-all-classes

  @classmethod
  def create_updated_stats(cls,
                           stats,
                           samples: jnp.ndarray,
                           *,
                           axis: Optional[Iterable[int]] = None,
                           paxis_name: Optional[str] = None,
                           alpha: Optional[float] = None,
                           mask: Optional[jnp.ndarray] = None,
                           exclude_zeros: bool = False):
    """Create a new Stats instance that represents the updated statistics.

    Since flax.struct.dataclass objects are frozen, this method creates a new
    instance of Stats with updated stats and returns it.

    Args:
      stats: A Stats dataclass object to be updated.
      samples: An array to update the current statistics with.
      axis: axis to average input samples over, e.g. to calculate stats per
        channel.
      paxis_name: the axis name used to combine batch statistics from multiple
        devices. See `jax.pmap` for a description of axis names.
      alpha: Smoothing parameter to use for moving average. If None, will use
        1/n, where n is the stat count.
      mask: Optional boolean tensor of the same shape as 'samples' specifying
        which values of 'samples' to use as part of the bounds calculation.
        'True' indicates the corresponding value from 'samples' should be used.
        If None, all values are used.
      exclude_zeros: Whether to exclude zeros in samples when computing
        statistics, e.g. when calculating mean absolute values.

    Returns:
      A new Stats instance with updated stats and count.
    """

    if mask is None:
      mask = jnp.full(samples.shape, True)
    shape_utils.assert_shapes_compatible(samples.shape, mask.shape)
    mask = jnp.broadcast_to(mask, samples.shape)
    if exclude_zeros:
      # Where samples are zero, set mask to False. This way they won't be
      # included in statistics.
      mask = mask & (samples != 0)

    def _moving_avg(old_avg: jnp.ndarray, new_val: jnp.ndarray,
                    masked_reduction_fn: Callable[..., jnp.ndarray]):
      masked_new_val_reduced = masked_reduction_fn(
          new_val, mask=mask, axis=axis, paxis_name=paxis_name, keepdims=True)
      valid_mask = jnp.isfinite(masked_new_val_reduced)
      # Only update average where means are valid, so set deltas corresponding
      # to invalid entries to 0.
      delta = jnp.where(valid_mask, masked_new_val_reduced - old_avg, 0)
      # TODO(lew): This is slightly incorrect, alpha should be proportional to
      # the mask size.
      new_avg = old_avg + alpha * delta
      return new_avg

    new_n = stats.n + 1
    if alpha is None:
      alpha = 1. / new_n

    new_mean = _moving_avg(stats.mean, samples, masked_reduction_fn=masked_mean)
    new_mean_abs = _moving_avg(
        stats.mean_abs, jnp.abs(samples), masked_reduction_fn=masked_mean)
    new_mean_sq = _moving_avg(
        stats.mean_sq, jnp.square(samples), masked_reduction_fn=masked_mean)
    new_mean_batch_minimum = _moving_avg(
        stats.mean_batch_minimum,
        samples,
        masked_reduction_fn=masked_mean_of_min)
    new_mean_batch_maximum = _moving_avg(
        stats.mean_batch_maximum,
        samples,
        masked_reduction_fn=masked_mean_of_max)
    return cls(
        n=new_n,
        mean=new_mean,
        mean_abs=new_mean_abs,
        mean_sq=new_mean_sq,
        mean_batch_minimum=new_mean_batch_minimum,
        mean_batch_maximum=new_mean_batch_maximum)  # pytype: disable=wrong-keyword-args  # trace-all-classes



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Numerics for int8, int4, binary and other integer types."""

from typing import Any
from aqt.jax.v2 import stochastic_rounding
from aqt.jax.v2 import utils
from aqt.jax.v2.numerics import numerics
from jax import lax
import jax.numpy as jnp


@utils.flax_slots_kw_only_dataclass
class IntSymmetric(numerics.AqtNumerics):
  """Symmetric numerics for sint8, sint4, binary, etc."""

  bits: int
  preserve_zero: bool
  # false = map max val on the end of the last bucket
  # true = map max val on the middle of the last
  preserve_max_val: bool
  # The quantized values are only guarenteed to be within the appropriate signed
  # int range if clip=True and round=True.
  clip: bool
  clip_gradient: bool
  round: bool
  noise_fn: None | stochastic_rounding.NoiseFn
  dtype: None | Any = None

  # pylint: disable=line-too-long
  # Verifying the correctness of these functions amounts to verifying this table:
  # if preserve_zero == F, zero might be rounded either to [-1, 0] bucket or to [0, 1] bucket
  # preserve_zero, preserve_max_val, 8b, 2b, 1b
  # F, F, 128.0, 2.0, 1.0  # bucket count is even; map onto the far edge of the last bucket
  # F, T, 127.5, 1.5, 0.5  # bucket count is even; map onto the center of the last bucket
  # T, F, 127.5, 1.5, 0.5  # bucket count is odd;  map onto the far edge of the last bucket
  # T, T, 127.0, 1.0, 0.0  # bucket count is odd;  map onto the center of the last bucket
  # pylint: enable=line-too-long

  def get_edge_of_last_int_bucket(self):
    ret = 2.0 ** (self.bits - 1)
    if self.preserve_zero:
      # Lose one bucket.
      ret -= 0.5
    return ret

  def get_center_of_last_int_bucket(self):
    return self.get_edge_of_last_int_bucket() - 0.5

  def get_quant_bound(self):
    if self.preserve_max_val:
      return self.get_center_of_last_int_bucket()
    else:
      return self.get_edge_of_last_int_bucket()

  def _get_fwd_clip_bound(self):
    # If we are not rounding, we just clip to bucket edges.
    fwd_clip_bound = self.get_edge_of_last_int_bucket()
    # If, after clip, we are rounding, we need to make sure that
    # we won't round values at the clip_bound away to the
    # non-existing bucket.
    if self.round:
      # Reducing fwd_clip_bound by any value in (0.0, 1.0) is correct.
      fwd_clip_bound -= 0.5
    return fwd_clip_bound

  def get_dtype(self):
    return self.dtype

  def vjp_fwd(self, x, context):
    """Forward pass."""
    res = (x,)
    input_dtype = x.dtype
    assert self.bits <= 22, 'Too many bits, float32 has less precision.'

    # Maybe noise
    if self.noise_fn:
      assert context.key is not None, (
          'noise_fn is set, requestic stochastic rounding, but RNG was not '
          'passed in Context.key'
      )
      x = (x + self.noise_fn(x.shape, context.key)).astype(input_dtype)

    if self.clip:
      fwd_clip_bound = self._get_fwd_clip_bound()
      x = jnp.clip(x, -fwd_clip_bound, fwd_clip_bound)

    # Maybe round
    if self.round:
      # TODO(lew): Have bucket centers at 2*k + 1, not at halves.
      round_to_halves = not self.preserve_zero
      if round_to_halves:
        x = jnp.floor(x) + 0.5
      else:
        x = lax.round(x, lax.RoundingMethod.TO_NEAREST_EVEN)

    # Maybe cast: return dtype is either int or the input dtype
    dtype = self.get_dtype()
    x = x.astype(dtype if dtype is not None else input_dtype)
    return x, res

  def vjp_bwd(self, res, grad):
    # Gradient of the clip function.
    # For boundary values we will have full gradient.
    # When using abs(max(x)) scaling, x is always in the interior, and the
    # gradient clip is always 1. So, we can always set clip_gradient to false.
    # However, other types of scaling may result in x being outside (i.e., there
    # is clipping). In that case it may be desirable to make the gradient zero.
    ret = grad
    if self.clip_gradient:
      (x,) = res
      clip_bound = self._get_fwd_clip_bound()
      ret *= (-clip_bound <= x) * (x <= clip_bound)
    return (ret, None)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base abstract class for all numerics."""
import abc

from aqt.jax.v2 import utils


@utils.flax_slots_kw_only_dataclass
class AqtNumerics(abc.ABC):
  """Abstract class for various quantization numerics."""

  @abc.abstractmethod
  def get_dtype(self):
    pass

  @abc.abstractmethod
  def get_quant_bound(self):
    """The width that the bound corresponds to in the quantization range.

    For symmetric scaling (relative to a fixed zero point), the bound represents
    the largest absolute value that should be representable in the input tensor,
    and the quant_bound represents where that corresponds to in the
    quantization range.

    It could be biggest value that can be represented by numerical format
    exactly (e.g. in case of int8, 127), or it could be edge of the last bucket
    (e.g. in case of int8, 127.5. In this case the largest representable
    absolute value will be slightly smaller than the bound).

    For asymmetric scaling, the bound corresponds to the width of the entire
    input tensor, and the quant_bound corresponds to the width of the entire
    quantization range (e.g. in case of int8, 255).
    """
    pass

  @abc.abstractmethod
  def vjp_fwd(self, x, context):
    pass

  @abc.abstractmethod
  def vjp_bwd(self, res, grad):
    pass



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generates ConfigDict instances for the Imagenet.

Most users will just create a ConfigDict instance with 'get_config' and then
override its parameters to specialize the configuration.
"""

from aqt.jax_legacy.utils import config_schema_utils
import ml_collections

float_ph = config_schema_utils.float_ph
int_ph = config_schema_utils.int_ph
str_ph = config_schema_utils.str_ph
bool_ph = config_schema_utils.bool_ph


def get_base_config(use_auto_acts: bool) -> ml_collections.ConfigDict:
  """Base ConfigDict for resnet, does not yet have fields for individual layers."""
  base_config = config_schema_utils.get_base_config(
      use_auto_acts, fp_quant=False)
  base_config.update({
      "base_learning_rate": float_ph(),
      "momentum": float_ph(),
      "model_hparams": {},
      "act_function": str_ph(),  # add a new field that controls act function
      "shortcut_ch_shrink_method": str_ph(),
      "shortcut_ch_expand_method": str_ph(),
      "shortcut_spatial_method": str_ph(),
      "lr_scheduler": {
          "warmup_epochs": int_ph(),
          "cooldown_epochs": int_ph(),
          "scheduler": str_ph(),  # "cosine", "linear", or "step" lr decay
          "num_epochs": int_ph(),
          "endlr": float_ph(),
          "knee_lr": float_ph(),
          "knee_epochs": int_ph(),
      },
      "optimizer": str_ph(),
      "adam": {
          "beta1": float_ph(),
          "beta2": float_ph()
      },
      "early_stop_steps": int_ph(),
      "weight_quant_start_step": int_ph(),
      "teacher_model": str_ph(),
      "is_teacher": bool_ph(),  # whether to train vanilla resnet or PokeBNN
      "seed": int_ph(),
  })
  if use_auto_acts:
    # config_schema_utils is shared by wmt. To not make other code libraries
    # affected, add the new bound coefficients here.
    base_config.quant_act.bounds.update({
        "fixed_bound": float_ph(),
        "cams_coeff": float_ph(),
        "cams_stddev_coeff": float_ph(),
        "mean_of_max_coeff": float_ph(),
        "use_old_code": bool_ph(),
    })

  base_config.dense_layer = config_schema_utils.get_dense_config(base_config)
  # TODO(b/179063860): The input distribution is an intrinsic model
  # property and shouldn't be part of the model configuration. Update
  # the hparam dataclasses to eliminate the input_distribution field and
  # then delete this.
  base_config.dense_layer.quant_act.input_distribution = "positive"
  base_config.conv = config_schema_utils.get_conv_config(base_config)
  base_config.residual = get_residual_config(base_config)
  # make the activation function in a residual block consistent with the global
  # option
  config_schema_utils.set_default_reference(
      base_config.residual,
      base_config,
      field=[
          "act_function", "shortcut_ch_shrink_method",
          "shortcut_ch_expand_method", "shortcut_spatial_method"
      ])
  return base_config


def get_residual_config(
    parent_config: ml_collections.ConfigDict) -> ml_collections.ConfigDict:
  """Creates ConfigDict corresponding to imagenet.models.ResidualBlock.HParams."""
  config = ml_collections.ConfigDict()
  config_schema_utils.set_default_reference(
      config,
      parent_config, ["conv_se", "conv_proj", "conv_1", "conv_2", "conv_3"],
      parent_field="conv")
  # TODO(b/179063860): The input distribution is an intrinsic model
  # property and shouldn't be part of the model configuration. Update
  # the hparam dataclasses to eliminate the input_distribution field and
  # then delete this.
  config.conv_proj.quant_act.input_distribution = "positive"
  config.conv_2.quant_act.input_distribution = "positive"
  config.conv_3.quant_act.input_distribution = "positive"
  # add a new field in a residual block that control the act function
  config.update({
      "act_function": str_ph(),
      "shortcut_ch_shrink_method": str_ph(),
      "shortcut_ch_expand_method": str_ph(),
      "shortcut_spatial_method": str_ph(),
  })

  config.lock()
  return config


def get_config(num_blocks: int,
               use_auto_acts: bool) -> ml_collections.ConfigDict:
  """Returns a ConfigDict instance for a Imagenet (Resnet50 and Resnet101).

  The ConfigDict is wired up so that changing a field at one level of the
  hierarchy changes the value of that field everywhere downstream in the
  hierarchy. For example, changing the top-level 'prec' parameter
  (eg, config.prec=4) will cause the precision of all layers to change.
  Changing the precision of a specific layer type
  (eg, config.residual_block.conv_1.weight_prec=4) will cause the weight prec
  of all conv_1 layers to change, overriding the value of the global
  config.prec value.

  See config_schema_test.test_schema_matches_expected to see the structure
  of the ConfigDict instance this will return.

  Args:
    num_blocks: Number of residual blocks in the architecture.
    use_auto_acts: Whether to use automatic clipping bounds for activations or
      fixed bounds. Unlike other properties of the configuration which can be
      overridden directly in the ConfigDict instance, this affects the immutable
      schema of the ConfigDict and so has to be specified before the ConfigDict
      is created.

  Returns:
    A ConfigDict instance which parallels the hierarchy of TrainingHParams.
  """
  base_config = get_base_config(use_auto_acts=use_auto_acts)
  model_hparams = base_config.model_hparams
  config_schema_utils.set_default_reference(model_hparams, base_config,
                                            "dense_layer")

  config_schema_utils.set_default_reference(
      model_hparams, base_config, "conv_init", parent_field="conv")
  model_hparams.residual_blocks = [
      config_schema_utils.make_reference(base_config, "residual")
      for _ in range(num_blocks)
  ]
  model_hparams.update({
      # Controls the number of parameters in the model by multiplying the number
      # of conv filters in each layer by this number.
      "filter_multiplier": float_ph(),
      "act_function": str_ph(),
      "se_ratio": float_ph(),
      "init_group": int_ph(),  # feature group in the second group conv layer
  })
  config_schema_utils.set_default_reference(
      model_hparams, base_config, "act_function", parent_field="act_function")

  base_config.lock()
  return base_config



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A sample to show how to quantize a model without calibration using AQT."""


from collections.abc import Sequence

from absl import app
from absl import flags
from aqt.jax.v2 import config as aqt_config
from aqt.jax.v2.examples.cnn import aqt_utils
from aqt.jax.v2.examples.cnn import cnn_model
from aqt.jax.v2.examples.cnn import model_utils
import jax

_WEIGHT_ONLY = flags.DEFINE_bool(
    'weight_only', False, 'Perform weight-only quantization.'
)


def run(train_ds: dict[str, jax.Array], test_ds: dict[str, jax.Array]) -> None:
  """Extracted main function for unit testing."""
  # 1. Train
  state = model_utils.create_train_state(
      jax.random.key(0), cnn_model.CNN(True), cnn_model.CNN(False)
  )
  state = model_utils.train_and_evaluate(
      num_epochs=1,
      workdir='/tmp/aqt_mnist_example',
      train_ds=train_ds,
      test_ds=test_ds,
      state=state,
  )

  # 2. Quantize and test
  aqt_cfg = aqt_config.default_unquantized_config()
  aqt_config.set_bits(
      aqt_cfg,
      fwd_lhs_bit=None if _WEIGHT_ONLY.value else 8,
      fwd_rhs_bit=8,
      dlhs_lhs_bit=None,
      dlhs_rhs_bit=None,
      drhs_lhs_bit=None,
      drhs_rhs_bit=None,
  )
  loss = aqt_utils.serve_quantized(
      cnn_model.CNN, test_ds, aqt_cfg, state.model_vars, act_calibrated=False
  )
  print('serve loss on sample ds: {}'.format(loss))


def main(argv: Sequence[str]) -> None:
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')

  train_ds, test_ds = cnn_model.get_datasets()
  run(train_ds, test_ds)


if __name__ == '__main__':
  app.run(main)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pokebnn.py."""

import hashlib
from absl.testing import absltest
from aqt.jax_legacy.jax.imagenet import hparams_config
from aqt.jax_legacy.jax.imagenet import models
from aqt.jax_legacy.jax.imagenet import train_utils as imagenet_train_utils
from aqt.jax_legacy.jax.imagenet.configs.pokebnn.pokebnn_config import get_config
from aqt.jax_legacy.utils import hparams_utils as os_hparams_utils
import jax
import jax.numpy as jnp


class PokeBNNTest(absltest.TestCase):

  def test_size_stats(self):
    config_dict = get_config().configs[0]
    hparams = os_hparams_utils.load_hparams_from_config_dict(
        hparams_config.TrainingHParams, models.ResNet.HParams, config_dict)

    model, _ = imagenet_train_utils.create_model(  #
        jax.random.PRNGKey(0),
        1,
        224,
        jnp.bfloat16,
        hparams=hparams.model_hparams,
        train=False,
        is_teacher=False)  # specify is_teacher=False for student

    # Regression test.
    stats = model.size_stats()
    full_str = str(stats).replace('}', '}\n')
    hashed = hashlib.sha224(full_str.encode()).hexdigest()

    self.assertEqual(
        hashed,
        'fd3bc7e2978b7260a5ed41dfcf3f8f783aa5d614c5e71d12ffb1186f',
        msg='Model size_stats changed. If expected, please update this hash.' +
        full_str)

    # Delete large fields and explicitely check a small core part.
    del stats['conv_stats']
    del stats['residual_params']
    self.assertEqual(
        stats, {
            'conv_ace': 4091469824.0,
            'dense_ace': 131072000,
            'batch_norm_sops': 17912832,
            'bprelu_sops': 9106944,
            'reshape_add_ch_shrink_sops': 5419008,
            'reshape_add_ch_shrink_mul_sops': 1705984,
            'reshape_add_avg_pool_sops': 7902720,
            'reshape_add_block_residual_sops': 5519360,
            'reshape_add_local_residual_sops': 8805888,
            'se_global_pool_add_sops': 8906240,
            'se_global_pool_mul_sops': 20672,
            'se_relu_sops': 2584,
            'se_relu6_sops': 22656,
            'se_mul_sops': 8805888,
            'pokebnn_global_pool_sops': 100352,
            'dense_stats': {
                'dense': {
                    'input_shape': 2048,
                    'input_prec': 8,
                    'weight_shape': (2048, 1000),
                    'weight_prec': 8,
                    'output_shape': 1000,
                    'dense_ace': 131072000
                }
            }
        })


if __name__ == '__main__':
  absltest.main()
